import json
import mock
import pytest
import requests
import requests.exceptions
import requests_mock

from pubtools._quay import quay_client, exceptions
from pubtools._quay.exceptions import ManifestNotFoundError


@mock.patch("pubtools._quay.quay_client.QuaySession")
def test_init(mock_session):
    client = quay_client.QuayClient("user", "pass", "stage.quay.io")

    assert client.username == "user"
    assert client.password == "pass"
    assert client.session == mock_session.return_value
    mock_session.assert_called_once_with(api="docker", hostname="stage.quay.io")


@mock.patch("pubtools._quay.quay_client.QuaySession")
def test_parse_image(mock_session):
    client = quay_client.QuayClient("user", "pass", "stage.quay.io")

    repo, ref = client._parse_and_validate_image_url("quay.io/name/image:1")
    assert repo == "name/image"
    assert ref == "1"

    repo, ref = client._parse_and_validate_image_url("quay.io/name2/image2@sha256:dfgdfg8df5g")
    assert repo == "name2/image2"
    assert ref == "sha256:dfgdfg8df5g"

    with pytest.raises(ValueError, match="Neither tag nor digest were found in the image"):
        client._parse_and_validate_image_url("quay.io/name/image")


@mock.patch("pubtools._quay.quay_client.QuaySession")
def test_authenticate_quay_header_error(mock_session):
    client = quay_client.QuayClient("user", "pass")

    header1 = {
        "Server": "nginx/1.12.1",
        "Date": "Tue, 02 Feb 2021 13:03:35 GMT",
        "Content-Type": "application/json",
        "Content-Length": "112",
        "Connection": "close",
        "Docker-Distribution-API-Version": "registry/2.0",
    }
    with pytest.raises(exceptions.RegistryAuthError, match="'WWW-Authenticate' is not in the.*"):
        client._authenticate_quay(header1)

    header2 = {
        "Server": "nginx/1.12.1",
        "Date": "Tue, 02 Feb 2021 13:03:35 GMT",
        "Content-Type": "application/json",
        "Content-Length": "112",
        "Connection": "close",
        "Docker-Distribution-API-Version": "registry/2.0",
        "WWW-Authenticate": 'Basic realm="https://quay.io/v2/auth",service="quay.io",'
        'scope="repository:namespace/some-repo:pull"',
    }
    with pytest.raises(exceptions.RegistryAuthError, match="Different than the Bearer.*"):
        client._authenticate_quay(header2)


@mock.patch("pubtools._quay.quay_client.QuaySession")
@mock.patch("pubtools._quay.quay_client.requests.Session")
def test_authenticate_quay_success(mock_session, mock_quay_session):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = {"token": "abcdef"}
    mock_get = mock.MagicMock()
    mock_get.return_value = mock_response
    mocked_session = mock.MagicMock()
    mocked_session.get = mock_get
    mock_session.return_value = mocked_session

    mocked_quay_session = mock.MagicMock()
    mock_quay_session.return_value = mocked_quay_session

    client = quay_client.QuayClient("user", "pass")
    header = {
        "Server": "nginx/1.12.1",
        "Date": "Tue, 02 Feb 2021 13:03:35 GMT",
        "Content-Type": "application/json",
        "Content-Length": "112",
        "Connection": "close",
        "Docker-Distribution-API-Version": "registry/2.0",
        "WWW-Authenticate": 'Bearer realm="https://quay.io/v2/auth",service="quay.io",'
        'scope="repository:namespace/some-repo:pull"',
    }
    client._authenticate_quay(header)

    mocked_session.get.assert_called_once_with(
        "https://quay.io/v2/auth",
        auth=("user", "pass"),
        params={"service": "quay.io", "scope": "repository:namespace/some-repo:pull"},
        timeout=10,
    )
    mocked_quay_session.set_auth_token.assert_called_once_with("abcdef")


@mock.patch("pubtools._quay.quay_client.QuaySession")
@mock.patch("pubtools._quay.quay_client.requests.Session")
def test_authenticate_quay_missing_token(mock_session, mock_quay_session):
    mock_response = mock.MagicMock()
    mock_response.json.return_value = {"bad_data": "abcdef"}
    mock_get = mock.MagicMock()
    mock_get.return_value = mock_response
    mocked_session = mock.MagicMock()
    mocked_session.get = mock_get
    mock_session.return_value = mocked_session

    mocked_quay_session = mock.MagicMock()
    mock_quay_session.return_value = mocked_quay_session

    client = quay_client.QuayClient("user", "pass")
    header = {
        "Server": "nginx/1.12.1",
        "Date": "Tue, 02 Feb 2021 13:03:35 GMT",
        "Content-Type": "application/json",
        "Content-Length": "112",
        "Connection": "close",
        "Docker-Distribution-API-Version": "registry/2.0",
        "WWW-Authenticate": 'Bearer realm="https://quay.io/v2/auth",service="quay.io",'
        'scope="repository:namespace/some-repo:pull"',
    }

    with pytest.raises(exceptions.RegistryAuthError, match="Authentication server response.*"):
        client._authenticate_quay(header)


@mock.patch("pubtools._quay.quay_client.QuayClient._authenticate_quay")
def test_request_quay_success(mock_authenticate):
    with requests_mock.Mocker() as m:
        m.get("https://quay.io/v2/get/data/1", text="data")

        client = quay_client.QuayClient("user", "pass")
        r = client._request_quay("GET", "get/data/1")

        assert r.text == "data"
        assert r.status_code == 200
        mock_authenticate.assert_not_called()


def test_request_quay_bad_status_code():
    with requests_mock.Mocker() as m:
        m.get("https://quay.io/v2/get/data/1", text="data", status_code=404)

        client = quay_client.QuayClient("user", "pass")
        with pytest.raises(requests.HTTPError, match="404 Client Error.*"):
            client._request_quay("GET", "get/data/1")


@mock.patch("pubtools._quay.quay_client.QuayClient._authenticate_quay")
def test_request_quay_authenticate_success(mock_authenticate):
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/get/data/1",
            [
                {"headers": {"some-header": "value"}, "status_code": 401},
                {"text": "data", "status_code": 200},
            ],
        )

        client = quay_client.QuayClient("user", "pass")
        r = client._request_quay("GET", "get/data/1")

        assert r.text == "data"
        assert r.status_code == 200
        mock_authenticate.assert_called_once_with({"some-header": "value"})


@mock.patch("pubtools._quay.quay_client.QuayClient._authenticate_quay")
def test_request_quay_authenticate_missing(mock_authenticate):
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/get/data/1",
            [
                {"headers": {"some-header": "value"}, "status_code": 401},
                {"text": "missing", "status_code": 404},
            ],
        )

        client = quay_client.QuayClient("user", "pass")

        with pytest.raises(requests.HTTPError, match="404 Client Error.*"):
            client._request_quay("GET", "get/data/1")
        mock_authenticate.assert_called_once_with({"some-header": "value"})


def test_get_manifest_list_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_ml = client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.list.v2+json",
        )
        assert m.call_count == 1

    assert ml == ret_ml


def test_get_manifest_list_raw_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=ml,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_ml = client.get_manifest(
            "quay.io/namespace/image:1",
            raw=True,
            media_type="application/vnd.docker.distribution.manifest.list.v2+json",
        )
        assert m.call_count == 1

    assert json.dumps(ml) == ret_ml


def test_get_manifest_list_wrong_type():
    manifest = {
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
        "size": 429,
        "digest": "sha256:6d5f4d65fg4d6f54g",
        "platform": {"architecture": "arm64", "os": "linux"},
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        with pytest.raises(exceptions.ManifestTypeError, match=".*doesn't have a .* manifest"):
            client.get_manifest(
                "quay.io/namespace/image:1",
                media_type="application/vnd.docker.distribution.manifest.list.v2+json",
            )
        assert m.call_count == 1


def test_get_manifest_success():
    manifest = {
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
        "size": 429,
        "digest": "sha256:6d5f4d65fg4d6f54g",
        "platform": {"architecture": "arm64", "os": "linux"},
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_manifest = client.get_manifest("quay.io/namespace/image:1")
        assert m.call_count == 2

    assert manifest == ret_manifest


def test_get_manifest_v2s2_second_attempt():
    v2s1_manifest = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 1,
        "signatures": [],
    }
    v2s2_manifest = {
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
        "size": 429,
        "digest": "sha256:6d5f4d65fg4d6f54g",
        "platform": {"architecture": "arm64", "os": "linux"},
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            [
                {
                    "headers": {
                        "Content-Type": "application/vnd.docker.distribution.manifest.v1+json"
                    },
                    "json": v2s1_manifest,
                },
                {
                    "headers": {
                        "Content-Type": "application/vnd.docker.distribution.manifest.v2+json"
                    },
                    "json": v2s2_manifest,
                },
            ],
        )

        client = quay_client.QuayClient("user", "pass")
        ret_manifest = client.get_manifest("quay.io/namespace/image:1")
        assert m.call_count == 2

    assert v2s2_manifest == ret_manifest


def test_get_manifest_accept_any():
    v2s1_manifest = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 1,
        "signatures": [],
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=v2s1_manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v1+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_manifest = client.get_manifest("quay.io/namespace/image:1")
        assert m.call_count == 5

    assert v2s1_manifest == ret_manifest


def test_get_v2s1_manifest():
    v2s1_manifest = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 1,
        "signatures": [],
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=v2s1_manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v1+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_manifest = client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.v1+json",
        )
        assert m.call_count == 1

    assert v2s1_manifest == ret_manifest


def test_get_v2s2_manifest():
    v2s2_manifest = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 2,
        "signatures": [],
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=v2s2_manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_manifest = client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.v2+json",
        )
        assert m.call_count == 1

    assert v2s2_manifest == ret_manifest


def test_get_v2s1_manifest_wrong_type():
    v2s2_manifest = {
        "mediaType": "application/vnd.docker.distribution.manifest.v1+json",
        "size": 429,
        "digest": "sha256:6d5f4d65fg4d6f54g",
        "platform": {"architecture": "arm64", "os": "linux"},
    }

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            json=v2s2_manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v1+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        with pytest.raises(
            exceptions.ManifestTypeError,
            match=r".*doesn't have a application/vnd\."
            r"docker\.distribution\.manifest\.v2\+json manifest",
        ):
            client.get_manifest(
                "quay.io/namespace/image:1",
                media_type="application/vnd.docker.distribution.manifest.v2+json",
            )
        assert m.call_count == 1


def test_get_v2s1_manifest_raw():
    v2s1_manifest_data = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 1,
        "signatures": [],
    }
    v2s1_manifest = json.dumps(v2s1_manifest_data, sort_keys=True, indent=4)

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            text=v2s1_manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v1+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        ret_manifest = client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.v1+json",
            raw=True,
        )
        assert m.call_count == 1

    assert v2s1_manifest == ret_manifest


def test_upload_manifest_list_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }

    with requests_mock.Mocker() as m:
        m.put("https://quay.io/v2/namespace/image/manifests/1", status_code=200)

        client = quay_client.QuayClient("user", "pass")
        client.upload_manifest(ml, "quay.io/namespace/image:1")
        assert m.call_count == 1
        assert m.request_history[0].json() == ml


def test_upload_raw_manifest_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }

    with requests_mock.Mocker() as m:
        m.put("https://quay.io/v2/namespace/image/manifests/1", status_code=200)

        client = quay_client.QuayClient("user", "pass")
        client.upload_manifest(json.dumps(ml), "quay.io/namespace/image:1", raw=True)
        assert m.call_count == 1
        assert m.request_history[0].json() == ml


def test_upload_manifest_list_failure():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }

    with requests_mock.Mocker() as m:
        m.put("https://quay.io/v2/namespace/image/manifests/1", status_code=400)

        client = quay_client.QuayClient("user", "pass")
        with pytest.raises(requests.HTTPError, match="400 Client Error.*"):
            client.upload_manifest(ml, "quay.io/namespace/image:1")
        assert m.call_count == 1


def test_get_repository_tags_no_pagination():
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/tags/list",
            json={"name": "namespace/image", "tags": ["1", "2", "3"]},
        )

        client = quay_client.QuayClient("user", "pass")
        data = client.get_repository_tags("namespace/image")
        assert data == {"name": "namespace/image", "tags": ["1", "2", "3"]}
        assert m.call_count == 1


def test_get_repository_tags_raw():
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/tags/list",
            json={"name": "namespace/image", "tags": ["1", "2", "3"]},
        )

        client = quay_client.QuayClient("user", "pass")
        data = client.get_repository_tags("namespace/image", "raw")
        assert data == json.dumps({"name": "namespace/image", "tags": ["1", "2", "3"]})
        assert m.call_count == 1


def test_get_repository_tags_pagination():
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/tags/list",
            json={"name": "namespace/image", "tags": ["1", "2", "3"]},
            headers={"Link": '</v2/namespace/image/tags/list/next-page-2>; rel="next"'},
        )
        m.get(
            "https://quay.io/v2/namespace/image/tags/list/next-page-2",
            json={"name": "namespace/image", "tags": ["4", "5", "6"]},
            headers={"Link": '</v2/namespace/image/tags/list/next-page-3>; rel="next"'},
        )
        m.get(
            "https://quay.io/v2/namespace/image/tags/list/next-page-3",
            json={"name": "namespace/image", "tags": ["7", "8", "9"]},
        )

        client = quay_client.QuayClient("user", "pass")
        data = client.get_repository_tags("namespace/image")
        assert data == {
            "name": "namespace/image",
            "tags": ["1", "2", "3", "4", "5", "6", "7", "8", "9"],
        }
        assert m.call_count == 3


def test_get_repository_tags_pagination_cannot_parse_url():
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/tags/list",
            json={"name": "namespace/image", "tags": ["1", "2", "3"]},
            headers={"Link": '/v2/namespace/image/tags/list/next-page-2; rel="next"'},
        )

        client = quay_client.QuayClient("user", "pass")
        with pytest.raises(ValueError, match="Could not extract next page URL.*"):
            client.get_repository_tags("namespace/image")

        assert m.call_count == 1


def test_get_manifest_digest():
    manifest = (
        '{"mediaType": "application/vnd.docker.distribution.manifest.v2+json",'
        ' "size": 429, "digest": "sha256:6d5f4d65fg4d6f54g",'
        ' "platform": {"architecture": "arm64", "os": "linux"}}'
    )

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            text=manifest,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        client = quay_client.QuayClient("user", "pass")
        digest = client.get_manifest_digest("quay.io/namespace/image:1")
        assert m.call_count == 2

        assert digest == "sha256:b9742c91f353022604e8ed4cf4ab1d688114fc4b133f0e11cbf7dd6272753ac8"


def test_get_manifest_digest_not_found():
    client = quay_client.QuayClient("user", "pass")
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            status_code=404,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        with pytest.raises(ManifestNotFoundError):
            client.get_manifest_digest("quay.io/namespace/image:1")


def test_get_manifest_digest_unkown_error():
    client = quay_client.QuayClient("user", "pass")
    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/namespace/image/manifests/1",
            status_code=500,
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        with pytest.raises(requests.exceptions.HTTPError):
            client.get_manifest_digest("quay.io/namespace/image:1")



import hashlib
import json
import logging
import requests
import re
from urllib3.util.retry import Retry
from urllib import request
import threading
from typing import Any, cast, Dict, List, Tuple, Optional, Union, Mapping

from .exceptions import ManifestTypeError, RegistryAuthError, ManifestNotFoundError
from .quay_session import QuaySession
from .types import ManifestList, Manifest

LOG = logging.getLogger("pubtools.quay")


class QuayClient:
    """Class for performing Docker HTTP API operations with the Quay registry."""

    MANIFEST_LIST_TYPE = "application/vnd.docker.distribution.manifest.list.v2+json"
    MANIFEST_V2S2_TYPE = "application/vnd.docker.distribution.manifest.v2+json"
    MANIFEST_V2S1_TYPE = "application/vnd.docker.distribution.manifest.v1+json"
    MANIFEST_OCI_LIST_TYPE = "application/vnd.oci.image.index.v1+json"
    MANIFEST_OCI_V2S2_TYPE = "application/vnd.oci.image.manifest.v1+json"

    def __init__(
        self, username: Optional[str], password: Optional[str], host: Optional[str] = None
    ) -> None:
        """
        Initialize.

        Args:
            username (str):
                Quay username.
            password (str):
                Quay password.
            host (str):
                Quay registry URL.
        """
        self.username = username
        self.password = password
        self.host = host
        self.thread_local = threading.local()

    @property
    def session(self) -> Union[Any, QuaySession]:
        """Create QuaySession object per thread."""
        if not hasattr(self.thread_local, "session"):
            self.thread_local.session = QuaySession(hostname=self.host, api="docker")
        return self.thread_local.session

    def get_manifest(
        self,
        image: str,
        raw: bool = False,
        media_type: Optional[str] = None,
        return_headers: bool = False,
    ) -> Union[ManifestList, Manifest, str, Tuple[str, Mapping[str, str]]]:
        """
        Get manifest of given media type.

        Args:
            image (str):
                Image for which to get the manifest list.
            raw (bool):
                Whether to return the manifest as raw JSON.
            manifest_list (bool):
                Whether to only return a manifest list and raise an exception otherwise.
            media_type (str):
                Can be application/vnd.docker.distribution.manifest.list.v2+json,
                application/vnd.docker.distribution.manifest.v2+json,
                application/vnd.docker.distribution.manifest.v1+json,
                application/vnd.oci.image.manifest.v1+json, application/vnd.oci.image.index.v1+json
                or None indicating which manifest type is requested. If it's None,
                manifest list is prefered, but if v2s2 is returned instead, v2s2
                is returned as final result. If neither is found, same order is attempted with
                OCI type images.
        Returns (dict|str):
            Image manifest
        Raises:
            ManifestTypeError:
                When the image doesn't return the requested manifest type.
            ValueError:
                If Manifest list and V2S1 manifest are requested at the same time.
        """
        repo, ref = self._parse_and_validate_image_url(image)
        endpoint = "{0}/manifests/{1}".format(repo, ref)

        if media_type:
            kwargs = {"headers": {"Accept": media_type}}
            response = self._request_quay("GET", endpoint, kwargs)

            # text/plain may be returned for V2S1 by our CDN
            if (
                response.headers["Content-Type"] != media_type
                and "text/plain" not in response.headers["Content-Type"]
            ):
                raise ManifestTypeError(
                    "Image {0} doesn't have a {1} manifest".format(image, media_type)
                )
            if raw:
                if not return_headers:
                    return str(response.text)
                else:
                    return (str(response.text), response.headers)
            else:
                return cast(ManifestList, response.json())

        # If type is not specified, try to get manifests in this order
        # If somehow none of these match, we'll accept whatever we got
        for manifest_type in (
            QuayClient.MANIFEST_LIST_TYPE,
            QuayClient.MANIFEST_V2S2_TYPE,
            QuayClient.MANIFEST_OCI_LIST_TYPE,
            QuayClient.MANIFEST_OCI_V2S2_TYPE,
            QuayClient.MANIFEST_V2S1_TYPE,
        ):
            kwargs = {"headers": {"Accept": manifest_type}}
            response = self._request_quay("GET", endpoint, kwargs)

            if response.headers["Content-Type"] == manifest_type:
                break

        if raw:
            return response.text
        else:
            return cast(ManifestList, response.json())

    def get_manifest_digest(self, image: str, media_type: Optional[str] = None) -> str:
        """
        Get manifest of the specified image and calculate its digest by hashing it.

        Args:
            image (str):
                Image address for which to calculate the digest.
            media_type (str):
                Type of manifest can be application/vnd.docker.distribution.manifest.v2+json
                or application/vnd.docker.distribution.manifest.v1+json
        Returns (str):
            Manifest digest of the image.
        """
        try:
            manifest = cast(str, self.get_manifest(image, raw=True, media_type=media_type))
        except requests.exceptions.HTTPError as exc:
            if exc.response.status_code == 404:
                raise ManifestNotFoundError()
            else:
                raise exc
        # SHA 256 is pretty much the standard for container images
        hasher = hashlib.sha256()
        hasher.update(manifest.encode("utf-8"))
        digest = hasher.hexdigest()

        return "sha256:{0}".format(digest)

    def upload_manifest(
        self, manifest: Union[ManifestList, str], image: str, raw: bool = False
    ) -> None:
        """
        Upload manifest to a specified image.

        All manifest types are supported (manifest, manifest list).

        Args:
            manifest (dict):
                Manifest to be uploaded.
            image (str):
                Image address to upload the manifest to.
            raw (bool):
                Whether the given manifest is a string (raw) or a Python dictionary
        """
        repo, ref = self._parse_and_validate_image_url(image)
        endpoint = "{0}/manifests/{1}".format(repo, ref)

        if raw:
            manifest_type = json.loads(cast(str, manifest)).get(
                "mediaType", QuayClient.MANIFEST_V2S1_TYPE
            )
            kwargs = {
                "headers": {"Content-Type": manifest_type},
                "data": cast(str, manifest),
            }
        else:
            manifest_type = cast(Manifest, manifest).get("mediaType", QuayClient.MANIFEST_V2S1_TYPE)
            kwargs = {
                "headers": {"Content-Type": manifest_type},
                "data": json.dumps(cast(ManifestList, manifest), sort_keys=True, indent=4),
            }
        self._request_quay("PUT", endpoint, kwargs)

    def get_repository_tags(
        self, repository: str, raw: bool = False
    ) -> Union[str, Dict[str, List[str]]]:
        """
        Get tags of a provided repository.

        Args:
            repository (str):
                Repository whose tags should be gathered (expected format namespce/repo).
            raw (bool):
                Whether the given manifest is a string (raw) or a Python dictionary
        Returns (list):
            Tags which the repository contains.
        """
        endpoint = "{0}/tags/list".format(repository)
        response = self._request_quay("GET", endpoint)
        tags = response.json()

        while "Link" in response.headers:
            # next page response has format '</v2/....>; rel="next"'
            matches = re.findall('</v2/(.+?)>; rel="next"', response.headers["Link"])
            if len(matches) != 1:
                raise ValueError(
                    "Could not extract next page URL from response '{0}'".format(
                        response.headers["Link"]
                    )
                )
            response = self._request_quay("GET", matches[0])
            tags["tags"].extend(response.json()["tags"])

        if raw:
            return json.dumps(tags)
        else:
            return cast(Dict[str, List[str]], tags)

    def _request_quay(
        self, method: str, endpoint: str, kwargs: Dict[Any, Any] = {}
    ) -> requests.Response:
        """
        Perform a Docker HTTP API request on Quay registry. Handle authentication.

        Args:
            method (str):
                REST API method of the request (GET, POST, PUT, DELETE).
            endpoint (str):
                Endpoint of the request.
            kwargs (dict):
                Optional arguments to add to the Request object.
        Returns (Response):
            Request library's Response object.
        Raises:
            HTTPError: When the request returned an error status.
        """
        r = self.session.request(method, endpoint, **kwargs)
        # 401 is tolerated as Bearer token might need to be generated
        if r.status_code >= 400 and r.status_code < 600 and r.status_code != 401:
            r.raise_for_status()
        if r.status_code == 401:
            LOG.debug("Unauthorized request, attempting to authenticate.")
            self._authenticate_quay(r.headers)
        else:
            return r

        r = self.session.request(method, endpoint, **kwargs)
        r.raise_for_status()

        return r

    def _authenticate_quay(self, headers: Union[Dict[Any, Any], Mapping[str, Any]]) -> None:
        """
        Attempt to perform an authentication with registry's authentication server.

        Once authentication is complete, add the token to the Session object.
        Specifics can be found at https://docs.docker.com/registry/spec/auth/token/

        Args:
            headers (dict):
                Headers of the 401 response received from the registry.
        Raises:
            RegistryAuthError:
                When there's an issue with the authentication procedure.
        """
        if "WWW-Authenticate" not in headers:
            raise RegistryAuthError(
                "'WWW-Authenticate' is not in the 401 response's header. "
                "Authentication cannot continue."
            )
        if "Bearer " not in headers["WWW-Authenticate"]:
            raise RegistryAuthError(
                "Different than the Bearer authentication type was requested. "
                "Only Bearer is supported."
            )

        # parse header to get a dictionary
        params = request.parse_keqv_list(
            request.parse_http_list(headers["WWW-Authenticate"][len("Bearer ") :])  # noqa: E203
        )
        host = params.pop("realm")
        session = requests.Session()
        retry = Retry(
            total=3,
            read=3,
            connect=3,
            backoff_factor=2,
            status_forcelist=set(range(500, 512)),
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        # Make an authentication request to the specified realm with the provided REST parameters.
        # Basic username + password authentication is expected.
        r = session.get(
            host, params=params, auth=(self.username or "", self.password or ""), timeout=10
        )
        r.raise_for_status()

        if "token" not in r.json():
            raise RegistryAuthError("Authentication server response doesn't contain a token.")
        self.session.set_auth_token(r.json()["token"])

    def _parse_and_validate_image_url(self, image: str) -> Tuple[str, str]:
        """
        Extract image repository + reference from an image and validate its data.

        Args:
            image (str):
                A quay.io image.
        Returns (str, str):
            Tuple containing image repository (without base URL) and reference.
        Raises:
            ValueError:
                If image doesn't contain the expected data.
        """
        url_parts = image.split("/")
        if "@" in url_parts[-1]:
            remainder, ref = url_parts[-1].split("@")
        elif ":" in url_parts[-1]:
            remainder, ref = url_parts[-1].split(":")
        else:
            raise ValueError("Neither tag nor digest were found in the image")

        # Skip base URL and get last part of URL without reference
        repo = "/".join(url_parts[1:-1] + [remainder])
        return (repo, ref)



import json
import logging
import mock
import pytest
import requests_mock
import requests

from pubtools._quay import untag_images
from .utils.misc import compare_logs

# flake8: noqa: E501


@mock.patch("pubtools._quay.untag_images.untag_images")
def test_arg_constructor_required_args(mock_untag_images):
    required_args = [
        "dummy",
        "--reference",
        "quay.io/repo/some-image:1",
        "--quay-api-token",
        "some-token",
    ]
    untag_images.untag_images_main(required_args)
    _, called_args = mock_untag_images.call_args

    assert called_args["references"] == ["quay.io/repo/some-image:1"]
    assert called_args["quay_api_token"] == "some-token"


@mock.patch.dict("os.environ", {"QUAY_PASSWORD": "robot_token", "QUAY_API_TOKEN": "api_token"})
@mock.patch("pubtools._quay.untag_images.untag_images")
def test_arg_constructor_all_args(mock_untag_images):
    all_args = [
        "dummy",
        "--reference",
        "quay.io/repo/some-image:1",
        "--remove-last",
        "--quay-user",
        "some_user",
    ]
    untag_images.untag_images_main(all_args)
    _, called_args = mock_untag_images.call_args

    assert called_args["references"] == ["quay.io/repo/some-image:1"]
    assert called_args["remove_last"] == True
    assert called_args["quay_user"] == "some_user"
    assert called_args["quay_password"] == "robot_token"
    assert called_args["quay_api_token"] == "api_token"


@mock.patch("pubtools._quay.untag_images.untag_images")
def test_args_missing_reference(mock_untag_images):
    wrong_args = [
        "dummy",
        "--quay-api-token",
        "some-token",
    ]

    with pytest.raises(SystemExit) as system_error:
        untag_images.untag_images_main(wrong_args)

    assert system_error.type == SystemExit
    assert system_error.value.code == 2
    mock_untag_images.assert_not_called()


@mock.patch("pubtools._quay.untag_images.untag_images")
def test_args_missing_api_token(mock_untag_images):
    wrong_args = [
        "dummy",
        "--reference",
        "quay.io/repo/some-image:1",
    ]

    with pytest.raises(ValueError, match="--quay-api-token must be specified"):
        untag_images.untag_images_main(wrong_args)

    mock_untag_images.assert_not_called()


@mock.patch("pubtools._quay.untag_images.ImageUntagger")
def test_args_incorrect_digest_reference(mock_image_untagger):
    wrong_args = [
        "dummy",
        "--reference",
        "quay.io/repo/some-image@sha256:s5df6sd5f",
        "--quay-api-token",
        "some-token",
    ]

    with pytest.raises(ValueError, match="All references must be specified via tag, not digest"):
        untag_images.untag_images_main(wrong_args)

    mock_image_untagger.assert_not_called()


@mock.patch("pubtools._quay.untag_images.ImageUntagger")
def test_args_missing_quay_credential(mock_image_untagger):
    wrong_args = [
        "dummy",
        "--reference",
        "quay.io/repo/some-image:1",
        "--quay-api-token",
        "some-token",
        "--quay-user",
        "some_user",
    ]

    with pytest.raises(ValueError, match="Both user and password must be.*"):
        untag_images.untag_images_main(wrong_args)

    mock_image_untagger.assert_not_called()


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_full_run_remove_last(
    mock_triangulate, manifest_list_data, v2s2_manifest_data, caplog, hookspy
):
    args = [
        "dummy",
        "--reference",
        "quay.io/name/repo1:1",
        "--reference",
        "quay.io/name/repo1:2",
        "--quay-api-token",
        "some-token",
        "--remove-last",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
    ]
    caplog.set_level(logging.INFO)
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/name/repo1/tags/list",
            json=repo_tags,
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/1",
            text=json.dumps(manifest_list_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/2",
            text=json.dumps(manifest_list_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/3",
            text=json.dumps(v2s2_manifest_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/4",
            text=json.dumps(v2s2_manifest_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.delete("https://quay.io/api/v1/repository/name/repo1/tag/1")
        m.delete("https://quay.io/api/v1/repository/name/repo1/tag/2")
        untag_images.untag_images_main(args)

        expected_lost_images = [
            "quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f",
            "quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
        ]

        assert m.call_count == 16

        expected_logs = [
            "Started untagging operation with the following references: .*quay.io/name/repo1:1.*quay.io/name/repo1:2.*",
            "Gathering tags and digests of repository 'name/repo1'",
            "Getting cosign images of 6 images",
            "0 cosign images were found for the 6 images",
            "Getting cosign images of 0 images",
            "Following images won't be referencable by tag: "
            ".*quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f.*"
            ".*quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee.*"
            ".*quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9.*"
            ".*quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c.*"
            ".*quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb.*"
            ".*quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd.*",
            "Removing tag '1' from repository 'name/repo1'",
            "Removing tag '2' from repository 'name/repo1'",
            "Untagging operation succeeded",
        ]
        compare_logs(caplog, expected_logs)

    assert hookspy == [
        ("task_start", {}),
        (
            "quay_images_untagged",
            {
                "lost_refs": [
                    "quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
                    "quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
                    "quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
                    "quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f",
                    "quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
                    "quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
                ],
                "untag_refs": ["quay.io/name/repo1:1", "quay.io/name/repo1:2"],
            },
        ),
        ("task_stop", {"failed": False}),
    ]


def test_full_run_no_lost_digests(manifest_list_data, v2s2_manifest_data, caplog):
    args = [
        "dummy",
        "--reference",
        "quay.io/name/repo1:1",
        "--quay-api-token",
        "some-token",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
    ]
    caplog.set_level(logging.INFO)
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4", "5"]}

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/name/repo1/tags/list",
            json=repo_tags,
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/1",
            text=json.dumps(manifest_list_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/2",
            text=json.dumps(manifest_list_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/3",
            text=json.dumps(v2s2_manifest_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/4",
            text=json.dumps(v2s2_manifest_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        response = mock.MagicMock()
        response.status_code = 404
        m.register_uri(
            "GET",
            "https://quay.io/v2/name/repo1/manifests/5",
            exc=requests.exceptions.HTTPError("not found", response=response),
        )
        m.delete("https://quay.io/api/v1/repository/name/repo1/tag/1")
        untag_images.untag_images_main(args)

        assert m.call_count == 15

        expected_logs = [
            "Started untagging operation with the following references: .*quay.io/name/repo1:1.*",
            "Gathering tags and digests of repository 'name/repo1'",
            "No images will be lost by this untagging operation",
            "Removing tag '1' from repository 'name/repo1'",
            "Untagging operation succeeded",
        ]
        compare_logs(caplog, expected_logs)


@mock.patch("pubtools._quay.image_untagger.SecurityManifestPusher.cosign_triangulate_image")
def test_full_run_last_error(mock_triangulate, manifest_list_data, v2s2_manifest_data, caplog):
    args = [
        "dummy",
        "--reference",
        "quay.io/name/repo1:1",
        "--reference",
        "quay.io/name/repo1:2",
        "--quay-api-token",
        "some-token",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
    ]
    caplog.set_level(logging.INFO)
    repo_tags = {"name": "repo1", "tags": ["1", "2", "3", "4"]}

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/name/repo1/tags/list",
            json=repo_tags,
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/1",
            text=json.dumps(manifest_list_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/2",
            text=json.dumps(manifest_list_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.list.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/3",
            text=json.dumps(v2s2_manifest_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )
        m.get(
            "https://quay.io/v2/name/repo1/manifests/4",
            text=json.dumps(v2s2_manifest_data, sort_keys=True),
            headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
        )

        expected_err_msg = (
            "Following images .*"
            ".*quay.io/name/repo1@sha256:836b8281def8a913eb3f1aeb4d12d372d77e11fb4bc5ebffe46a55552af5fc1f.*"
            ".*quay.io/name/repo1@sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee.*"
            ".*quay.io/name/repo1@sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9.*"
            ".*quay.io/name/repo1@sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c.*"
            ".*quay.io/name/repo1@sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb.*"
            ".*quay.io/name/repo1@sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd.*"
        )

        with pytest.raises(ValueError, match=expected_err_msg):
            untag_images.untag_images_main(args)

        assert m.call_count == 14

        expected_logs = [
            "Started untagging operation with the following references: .*quay.io/name/repo1:1.*quay.io/name/repo1:2.*",
            "Gathering tags and digests of repository 'name/repo1'",
            "Getting cosign images of 6 images",
            "0 cosign images were found for the 6 images",
            "Getting cosign images of 0 images",
        ]
        compare_logs(caplog, expected_logs)


def test_full_run_get_manifest_error():
    args = [
        "dummy",
        "--reference",
        "quay.io/name/repo1:1",
        "--quay-api-token",
        "some-token",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
    ]
    repo_tags = {"name": "repo1", "tags": ["1"]}

    with requests_mock.Mocker() as m:
        m.get(
            "https://quay.io/v2/name/repo1/tags/list",
            json=repo_tags,
        )
        response = mock.MagicMock()
        response.status_code = 500
        m.register_uri(
            "GET",
            "https://quay.io/v2/name/repo1/manifests/1",
            exc=requests.exceptions.HTTPError("server error", response=response),
        )
        with pytest.raises(requests.exceptions.HTTPError, match="server error"):
            untag_images.untag_images_main(args)



import mock
import pytest

from pubtools._quay import remove_repo

from tests.utils.misc import GetManifestSideEffect


@mock.patch("pubtools._quay.remove_repo.remove_repositories")
def test_arg_constructor_required_args(mock_remove_repositories):
    required_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]
    remove_repo.remove_repositories_main(required_args)
    repositories, settings = mock_remove_repositories.call_args[0]
    assert repositories == "namespace/image"
    assert settings["quay_org"] == "quay-organization"
    assert settings["quay_user"] == "some-user"
    assert settings["quay_password"] == "some-password"
    assert settings["quay_api_token"] == "some-token"
    assert settings["pyxis_server"] == "pyxis-url.com"
    assert settings["pyxis_ssl_crtfile"] == "/path/to/file.crt"
    assert settings["pyxis_ssl_keyfile"] == "/path/to/file.key"
    assert settings["quay_host"] == "quay-host.com"


@mock.patch.dict("os.environ", {"QUAY_API_TOKEN": "api_token", "QUAY_PASSWORD": "some-password"})
@mock.patch("pubtools._quay.remove_repo.remove_repositories")
def test_arg_constructor_all_args(mock_remove_repositories):
    all_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]
    remove_repo.remove_repositories_main(all_args)
    repositories, called_kw_args = mock_remove_repositories.call_args[0]
    assert repositories == "namespace/image"
    assert called_kw_args["quay_org"] == "quay-organization"
    assert called_kw_args["quay_user"] == "some-user"
    assert called_kw_args["quay_password"] == "some-password"
    assert called_kw_args["pyxis_server"] == "pyxis-url.com"
    assert called_kw_args["pyxis_ssl_crtfile"] == "/path/to/file.crt"
    assert called_kw_args["pyxis_ssl_keyfile"] == "/path/to/file.key"
    assert called_kw_args["quay_api_token"] == "api_token"
    assert called_kw_args["quay_host"] == "quay-host.com"


@mock.patch("pubtools._quay.remove_repo.remove_repositories")
def test_args_missing_repository(mock_remove_repositories):
    wrong_args = [
        "dummy",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]

    with pytest.raises(SystemExit) as system_error:
        remove_repo.remove_repositories_main(wrong_args)

    assert system_error.type == SystemExit
    assert system_error.value.code == 2
    mock_remove_repositories.assert_not_called()


@mock.patch("pubtools._quay.remove_repo.remove_repositories")
def test_args_missing_api_token(mock_remove_repositories):
    wrong_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]

    with pytest.raises(ValueError, match="--quay-api-token must be specified"):
        remove_repo.remove_repositories_main(wrong_args)

    mock_remove_repositories.assert_not_called()


@mock.patch("pubtools._quay.remove_repo.remove_repositories")
def test_args_missing_quay_password(mock_remove_repositories):
    wrong_args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--quay-host",
        "quay-host.com",
    ]

    with pytest.raises(ValueError, match="--quay-password must be specified"):
        remove_repo.remove_repositories_main(wrong_args)

    mock_remove_repositories.assert_not_called()


@mock.patch("pubtools._quay.remove_repo.QuayClient")
@mock.patch("pubtools._quay.remove_repo.QuayApiClient")
def test_run(
    mock_quay_api_client,
    mock_quay_client,
    hookspy,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_remove_signatures,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    args = [
        "dummy",
        "--repositories",
        "namespace/image",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--signers",
        "msg_signer",
        "--quay-host",
        "quay-host.com",
    ]
    mock_delete_repo = mock.MagicMock()
    mock_quay_api_client.return_value.delete_repository = mock_delete_repo

    mock_get_repo_tags = mock.MagicMock()
    mock_get_repo_tags.return_value = {"tags": ["1", "2"]}
    mock_quay_client.return_value.get_repository_tags = mock_get_repo_tags

    # def get_manifest_side_effect(image, raw=False, media_type=False):
    #     if media_type == "application/vnd.docker.distribution.manifest.list.v2+json":
    #         content = src_manifest_list
    #     else:
    #         content = v2s1_manifest
    #     return json.dumps(content) if raw else content

    mock_quay_client.return_value.get_manifest.side_effect = GetManifestSideEffect(
        v2s1_manifest, src_manifest_list, call_limit=5
    )
    signer_wrapper_run_entry_point.return_value = [
        {
            "_id": 1,
            "manifest_digest": "sha256:3333333333",
            "reference": "some-registry.com/namespace/image:1",
            "sig_key_id": "key",
            "repository": "namespace/image",
        }
    ]
    remove_repo.remove_repositories_main(args)

    mock_quay_api_client.assert_called_once_with("some-token")
    signer_wrapper_remove_signatures.assert_called_once_with([1])
    mock_delete_repo.assert_called_once_with("quay-organization/namespace----image")

    assert hookspy == [
        ("task_start", {}),
        ("get_cert_key_paths", {"server_url": "pyxis-url.com"}),
        ("quay_repositories_removed", {"repository_ids": ["namespace/image"]}),
        ("task_stop", {"failed": False}),
    ]


@mock.patch("pubtools._quay.remove_repo.QuayClient")
@mock.patch("pubtools._quay.remove_repo.QuayApiClient")
def test_run_multiple_repos(
    mock_quay_api_client,
    mock_quay_client,
    hookspy,
    src_manifest_list,
    v2s1_manifest,
    fake_cert_key_paths,
    signer_wrapper_remove_signatures,
    signer_wrapper_entry_point,
    signer_wrapper_run_entry_point,
):
    args = [
        "dummy",
        "--repositories",
        "namespace/image,namespace/image2",
        "--quay-org",
        "quay-organization",
        "--quay-user",
        "some-user",
        "--quay-password",
        "some-password",
        "--quay-api-token",
        "some-token",
        "--pyxis-server",
        "pyxis-url.com",
        "--pyxis-ssl-crtfile",
        "/path/to/file.crt",
        "--pyxis-ssl-keyfile",
        "/path/to/file.key",
        "--signers",
        "msg_signer",
        "--quay-host",
        "quay-host.com",
    ]
    mock_delete_repo = mock.MagicMock()
    mock_quay_api_client.return_value.delete_repository = mock_delete_repo

    mock_get_repo_tags = mock.MagicMock()
    mock_get_repo_tags.return_value = {"tags": ["1", "2"]}
    mock_quay_client.return_value.get_repository_tags = mock_get_repo_tags

    # def get_manifest_side_effect(image, raw=False, media_type=False):
    #     if media_type == "application/vnd.docker.distribution.manifest.list.v2+json":
    #         content = src_manifest_list
    #     else:
    #         content = v2s1_manifest
    #     return json.dumps(content) if raw else content

    mock_quay_client.return_value.get_manifest.side_effect = GetManifestSideEffect(
        v2s1_manifest, src_manifest_list, call_limit=30
    )
    signer_wrapper_run_entry_point.return_value = [
        {
            "_id": 1,
            "manifest_digest": "sha256:3333333333",
            "reference": "some-registry.com/namespace/image2:1",
            "sig_key_id": "key",
            "repository": "namespace/image2",
        }
    ]

    remove_repo.remove_repositories_main(args)

    mock_quay_api_client.assert_called_once_with("some-token")
    signer_wrapper_remove_signatures.assert_called_once_with([1])
    assert mock_delete_repo.call_count == 2
    assert mock_delete_repo.call_args_list[0] == mock.call("quay-organization/namespace----image")
    assert mock_delete_repo.call_args_list[1] == mock.call("quay-organization/namespace----image2")



import mock
import pytest

from pubtools._quay import tag_images
from pubtools._quay.utils import misc


def test_argument_groups(capsys):
    args = {
        ("--arg1",): {
            "group": "Group 1",
            "help": "Argument 1",
            "required": True,
            "type": str,
        },
        ("--arg2",): {
            "group": "Group 1",
            "help": "Argument 2",
            "required": True,
            "type": str,
        },
        ("--arg3",): {
            "group": "Group 2",
            "help": "Argument 3",
            "required": True,
            "type": str,
        },
        ("--arg4",): {
            "group": "Group 2",
            "help": "Argument 4",
            "required": True,
            "type": str,
        },
    }

    parser = misc.setup_arg_parser(args)
    parser.print_help()
    out, _ = capsys.readouterr()

    assert "Group 1:" in out
    assert "Group 2:" in out


@mock.patch("pubtools._quay.tag_images.tag_images")
def test_arg_parser_required_args(mock_tag_images):
    required_args = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
    ]
    tag_images.tag_images_main(required_args)

    assert mock_tag_images.call_args == mock.call(
        source_ref="quay.io/repo/souce-image:1",
        all_arch=False,
        quay_user=None,
        quay_password=None,
        source_quay_host=None,
        source_quay_user=None,
        source_quay_password=None,
        remote_exec=False,
        ssh_remote_host=None,
        ssh_remote_host_port=None,
        ssh_reject_unknown_host=False,
        ssh_username=None,
        ssh_password=None,
        ssh_key_filename=None,
        container_exec=False,
        container_image=None,
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username=None,
        registry_password=None,
        dest_refs=["quay.io/repo/target-image:1"],
    )


@mock.patch("pubtools._quay.tag_images.tag_images")
def test_arg_parser_full_args(mock_tag_images):
    full_args = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--source-quay-host",
        "host_of_source",
        "--source-quay-user",
        "robot_user_for_source",
        "--source-quay-password",
        "robot_token_for_source",
        "--quay-user",
        "robot_user",
        "--quay-password",
        "robot_token",
        "--remote-exec",
        "--ssh-remote-host",
        "127.0.0.1",
        "--ssh-remote-host-port",
        "5000",
        "--ssh-reject-unknown-host",
        "--ssh-username",
        "dummy",
        "--ssh-password",
        "123456",
        "--ssh-key-filename",
        "/path/to/file.key",
        "--container-exec",
        "--container-image",
        "quay.io/namespace/image:1",
        "--docker-url",
        "https://some-url.com",
        "--docker-timeout",
        "120",
        "--docker-verify-tls",
        "--docker-cert-path",
        "/some/path",
        "--registry-username",
        "registry_user",
        "--registry-password",
        "registry_passwd",
    ]
    tag_images.tag_images_main(full_args)

    assert mock_tag_images.call_args == mock.call(
        source_ref="quay.io/repo/souce-image:1",
        all_arch=False,
        quay_user="robot_user",
        quay_password="robot_token",
        source_quay_host="host_of_source",
        source_quay_user="robot_user_for_source",
        source_quay_password="robot_token_for_source",
        remote_exec=True,
        ssh_remote_host="127.0.0.1",
        ssh_remote_host_port=5000,
        ssh_reject_unknown_host=True,
        ssh_username="dummy",
        ssh_password="123456",
        ssh_key_filename="/path/to/file.key",
        container_exec=True,
        container_image="quay.io/namespace/image:1",
        docker_url="https://some-url.com",
        docker_timeout="120",
        docker_verify_tls=True,
        docker_cert_path="/some/path",
        registry_username="registry_user",
        registry_password="registry_passwd",
        dest_refs=["quay.io/repo/target-image:1"],
    )


@mock.patch("pubtools._quay.tag_images.tag_images")
def test_arg_parser_multiple_args(mock_tag_images):
    multi_args = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:2",
    ]
    tag_images.tag_images_main(multi_args)

    assert mock_tag_images.call_args == mock.call(
        source_ref="quay.io/repo/souce-image:1",
        all_arch=False,
        quay_user=None,
        quay_password=None,
        source_quay_host=None,
        source_quay_user=None,
        source_quay_password=None,
        remote_exec=False,
        ssh_remote_host=None,
        ssh_remote_host_port=None,
        ssh_reject_unknown_host=False,
        ssh_username=None,
        ssh_password=None,
        ssh_key_filename=None,
        container_exec=False,
        container_image=None,
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username=None,
        registry_password=None,
        dest_refs=["quay.io/repo/target-image:1", "quay.io/repo/target-image:2"],
    )


@mock.patch("pubtools._quay.tag_images.tag_images")
def test_arg_parser_required_missing_source(mock_tag_images):
    missing_source = ["dummy", "--dest-ref", "quay.io/repo/target-image:1"]

    with pytest.raises(SystemExit) as system_error:
        tag_images.tag_images_main(missing_source)

    assert system_error.type == SystemExit
    assert system_error.value.code == 2
    mock_tag_images.assert_not_called()


@mock.patch("pubtools._quay.tag_images.tag_images")
def test_arg_parser_required_missing_dest(mock_tag_images):
    missing_dest = ["dummy", "--source-ref", "quay.io/repo/souce-image:1"]

    with pytest.raises(SystemExit) as system_error:
        tag_images.tag_images_main(missing_dest)

    assert system_error.type == SystemExit
    assert system_error.value.code == 2
    mock_tag_images.assert_not_called()


def test_arg_parser_missing_hostname():
    missing_hostname = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--remote-exec",
    ]

    with pytest.raises(ValueError, match="Remote host is missing.*"):
        tag_images.tag_images_main(missing_hostname)


def test_arg_parser_missing_quay_user_or_password():
    missing_password = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--quay-user",
        "robot_user",
    ]

    with pytest.raises(ValueError, match="Both user and password must be present.*"):
        tag_images.tag_images_main(missing_password)

    missing_user = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--quay-password",
        "robot_token",
    ]

    with pytest.raises(ValueError, match="Both user and password must be present.*"):
        tag_images.tag_images_main(missing_user)

    missing_source_password = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--quay-user",
        "robot_user",
        "--quay-password",
        "robot_token",
        "--source-quay-user",
        "robot_user_for_source",
    ]

    with pytest.raises(ValueError, match="Both source quay user and password must be present.*"):
        tag_images.tag_images_main(missing_source_password)

    missing_source_user = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--quay-user",
        "robot_user",
        "--quay-password",
        "robot_token",
        "--source-quay-password",
        "robot_token_for_source",
    ]

    with pytest.raises(ValueError, match="Both source quay user and password must be present.*"):
        tag_images.tag_images_main(missing_source_user)


def test_arg_parser_missing_container_image():
    missing_umb_url = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--container-exec",
    ]

    with pytest.raises(ValueError, match="Container image is missing when.*"):
        tag_images.tag_images_main(missing_umb_url)


@mock.patch.dict(
    "os.environ",
    {
        "SOURCE_QUAY_PASSWORD": "robot_token_for_source",
        "QUAY_PASSWORD": "robot_token",
        "SSH_PASSWORD": "123456",
        "REGISTRY_PASSWORD": "registry_passwd",
    },
)
@mock.patch("pubtools._quay.tag_images.tag_images")
def test_arg_parser_env_variables(mock_tag_images):
    full_args = [
        "dummy",
        "--source-ref",
        "quay.io/repo/souce-image:1",
        "--dest-ref",
        "quay.io/repo/target-image:1",
        "--source-quay-user",
        "robot_user_for_source",
        "--quay-user",
        "robot_user",
        "--remote-exec",
        "--ssh-remote-host",
        "127.0.0.1",
        "--ssh-reject-unknown-host",
        "--ssh-username",
        "dummy",
        "--ssh-key-filename",
        "/path/to/file.key",
        "--registry-username",
        "registry_user",
    ]
    tag_images.tag_images_main(full_args)

    assert mock_tag_images.call_args == mock.call(
        source_ref="quay.io/repo/souce-image:1",
        all_arch=False,
        quay_user="robot_user",
        quay_password="robot_token",
        source_quay_host=None,
        source_quay_user="robot_user_for_source",
        source_quay_password="robot_token_for_source",
        remote_exec=True,
        ssh_remote_host="127.0.0.1",
        ssh_remote_host_port=None,
        ssh_reject_unknown_host=True,
        ssh_username="dummy",
        ssh_password="123456",
        ssh_key_filename="/path/to/file.key",
        container_exec=False,
        container_image=None,
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username="registry_user",
        registry_password="registry_passwd",
        dest_refs=["quay.io/repo/target-image:1"],
    )



import functools
import logging
import argparse
from typing import Any, Optional, List, Dict, cast

from pubtools.pluggy import pm, task_context

from .utils.misc import setup_arg_parser, add_args_env_variables
from .command_executor import LocalExecutor, RemoteExecutor, ContainerExecutor

LOG = logging.getLogger("pubtools.quay")

TAG_IMAGES_ARGS = {
    ("--source-ref",): {
        "help": "Source image reference.",
        "required": True,
        "type": str,
    },
    ("--dest-ref",): {
        "help": "Destination image reference. Multiple can be specified.",
        "required": True,
        "type": str,
        "action": "append",
    },
    ("--all-arch",): {
        "help": "Flag of whether to copy all architectures of an image (if multiatch image)",
        "required": False,
        "type": bool,
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": False,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
    ("--source-quay-host",): {
        "help": "Host of source_ref.",
        "required": False,
        "type": str,
    },
    ("--source-quay-user",): {
        "help": "Username for source_ref registry login.",
        "required": False,
        "type": str,
    },
    ("--source-quay-password",): {
        "help": "Password for source_ref registry. Can be specified by env "
        "variable SOURCE_QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "SOURCE_QUAY_PASSWORD",
    },
    ("--remote-exec",): {
        "help": "Flag of whether the commands should be executed on a remote server.",
        "required": False,
        "type": bool,
    },
    ("--ssh-remote-host",): {
        "help": "Hostname for remote execution.",
        "required": False,
        "type": str,
    },
    ("--ssh-remote-host-port",): {
        "help": "Port of the remote host",
        "required": False,
        "type": int,
    },
    ("--ssh-reject-unknown-host",): {
        "help": "Flag of whether to reject an SSH host when it's not found among known hosts.",
        "required": False,
        "type": bool,
    },
    ("--ssh-username",): {
        "help": "Username for SSH connection. Defaults to local username.",
        "required": False,
        "type": str,
    },
    ("--ssh-password",): {
        "help": "Password for SSH. Will only be used if key-based validation is not available. "
        "Can be specified by env variable SSH_PASSWORD",
        "required": False,
        "type": str,
        "env_variable": "SSH_PASSWORD",
    },
    ("--ssh-key-filename",): {
        "help": "Path to the private key file for SSH authentication.",
        "required": False,
        "type": str,
    },
    ("--container-exec",): {
        "help": "Whether to execute the commands in a Docker container.",
        "required": False,
        "type": bool,
    },
    ("--container-image",): {
        "help": "Path to the container image in which to execute the commands. Must be "
        "downloadable without extra permissions.",
        "required": False,
        "type": str,
    },
    ("--docker-url",): {
        "help": "URL of the docker client that should run the container. Local socket by default.",
        "required": False,
        "type": str,
        "default": "unix://var/run/docker.sock",
    },
    ("--docker-timeout",): {
        "help": "Timeout for executing Docker commands. Disabled by default.",
        "required": False,
        "type": str,
    },
    ("--docker-verify-tls",): {
        "help": "Whether to perform TLS verification with the Docker client. Disabled by default.",
        "required": False,
        "type": bool,
    },
    ("--docker-cert-path",): {
        "help": "Path to Docker certificates for TLS authentication. '~/.docker' by default.",
        "required": False,
        "type": str,
    },
    ("--registry-username",): {
        "help": "Username to login to registry containing the specified image. If not provided, "
        "login will be assumed to not be needed.",
        "required": False,
        "type": str,
    },
    ("--registry-password",): {
        "help": "Password to login to registry containing the specified image. If not provided, "
        "login will be assumed to not be needed. "
        "Can be specified by env variable REGISTRY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "REGISTRY_PASSWORD",
    },
}


def construct_kwargs(args: argparse.Namespace) -> Dict[Any, Any]:
    """
    Construct a kwargs dictionary based on the entered command line arguments.

    Args:
        args (argparse.Namespace):
            Parsed command line arguments.
    Returns (dict):
        Keyword arguments for the 'tag_images' function.
    """
    kwargs = args.__dict__

    # in args.__dict__ unspecified bool values have 'None' instead of 'False'
    for name, attributes in TAG_IMAGES_ARGS.items():
        if attributes["type"] is bool:
            bool_var = name[0].lstrip("-").replace("-", "_")
            if kwargs[bool_var] is None:
                kwargs[bool_var] = False

    # some exceptions have to be remapped
    kwargs["dest_refs"] = kwargs.pop("dest_ref")

    return kwargs


def tag_images(
    source_ref: str,
    dest_refs: List[str],
    all_arch: bool = False,
    quay_user: Optional[str] = None,
    quay_password: Optional[str] = None,
    source_quay_host: Optional[str] = None,
    source_quay_user: Optional[str] = None,
    source_quay_password: Optional[str] = None,
    remote_exec: bool = False,
    ssh_remote_host: Optional[str] = None,
    ssh_remote_host_port: Optional[str] = None,
    ssh_reject_unknown_host: bool = False,
    ssh_username: Optional[str] = None,
    ssh_password: Optional[str] = None,
    ssh_key_filename: Optional[str] = None,
    container_exec: bool = False,
    container_image: Optional[str] = None,
    docker_url: str = "unix://var/run/docker.sock",
    docker_timeout: Optional[int] = None,
    docker_verify_tls: bool = False,
    docker_cert_path: Optional[str] = None,
    registry_username: Optional[str] = None,
    registry_password: Optional[str] = None,
) -> None:
    """
    Tag images in Quay.

    Args:
        source_ref (str):
            Source image reference.
        dest_refs ([str]):
            List of destination image references.
        all_arch (bool):
            Whether to copy all architectures.
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
        source_quay_host (str):
            Host of source ref.
        source_quay_user (str):
            Quay username for Docker HTTP API for the source ref.
        source_quay_password (str):
            Quay password for Docker HTTP API for the source ref.
        remote_exec (bool):
            Whether to execute the command remotely. Takes precedence over container_exec.
        ssh_remote_host (str):
            Hostname for remote execution.
        ssh_remote_host_port (str):
            Port of the remote host.
        ssh_reject_unknown_host (bool):
            whether to reject an SSH host when it's not found among known hosts.
        ssh_username (str):
            Username for SSH connection. Defaults to local username.
        ssh_password (str):
            Password for SSH. Will only be used if key-based validation is not available.
        ssh_key_filename (str):
            Path to the private key file for SSH authentication.
        container_exec (bool):
            Whether to execute the commands in a Docker container.
        container_image (str):
            Path to the container image in which to execute the commands. Must be downloadable
            without extra permissions.
        docker_url (str):
            URL of the docker client that should run the container. Local socket by default.
        docker_timeout (int):
            Timeout for executing Docker commands. Disabled by default.
        docker_verify_tls (bool):
            Whether to perform TLS verification with the Docker client. Disabled by default.
        docker_cert_path (str):
            Path to Docker certificates for TLS authentication. '~/.docker' by default.
        registry_username (str):
            Username to login to registry containing the specified image. If not provided,
            login will be assumed to not be needed.
        registry_password (str):
            Password to login to registry containing the specified image. If not provided,
            login will be assumed to not be needed.
    """
    verify_tag_images_args(
        quay_user,
        quay_password,
        source_quay_user,
        source_quay_password,
        remote_exec,
        ssh_remote_host,
        container_exec,
        container_image,
    )

    if remote_exec:
        accept_host = not ssh_reject_unknown_host if ssh_reject_unknown_host else True
        executor_class: (
            functools.partial[RemoteExecutor]
            | functools.partial[ContainerExecutor]
            | functools.partial[LocalExecutor]
        ) = functools.partial(
            RemoteExecutor,
            cast(str, ssh_remote_host),
            ssh_username,
            ssh_key_filename,
            ssh_password,
            cast(int, ssh_remote_host_port),
            accept_host,
        )
    elif container_exec:
        if isinstance(docker_timeout, str):
            docker_timeout = int(docker_timeout)
        executor_class = functools.partial(
            ContainerExecutor,
            cast(str, container_image),
            docker_url,
            docker_timeout,
            docker_verify_tls,
            docker_cert_path,
            registry_username,
            registry_password,
        )
    else:
        executor_class = functools.partial(LocalExecutor)

    with executor_class() as executor:
        dest_host = dest_refs[0].split("/", 1)[0]
        executor.skopeo_login(dest_host, quay_user, quay_password)
        if source_quay_host and source_quay_user and source_quay_password:
            executor.skopeo_login(source_quay_host, source_quay_user, source_quay_password)
        executor.tag_images(source_ref, dest_refs, all_arch)

    pm.hook.quay_images_tagged(source_ref=source_ref, dest_refs=sorted(dest_refs))


def verify_tag_images_args(
    quay_user: Optional[str],
    quay_password: Optional[str],
    source_quay_user: Optional[str],
    source_quay_password: Optional[str],
    remote_exec: bool,
    ssh_remote_host: Optional[str],
    container_exec: bool,
    container_image: Optional[str],
) -> None:
    """Verify the presence of input parameters."""
    if remote_exec:
        if not ssh_remote_host:
            raise ValueError("Remote host is missing when remote execution was specified.")

    if container_exec:
        if not container_image:
            raise ValueError("Container image is missing when container execution was specified.")

    if (quay_user and not quay_password) or (quay_password and not quay_user):
        raise ValueError("Both user and password must be present when attempting to log in.")

    if (source_quay_user and not source_quay_password) or (
        source_quay_password and not source_quay_user
    ):
        raise ValueError(
            "Both source quay user and password must be present when attempting to log in."
        )


def setup_args() -> argparse.ArgumentParser:
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(TAG_IMAGES_ARGS)


def tag_images_main(sysargs: Optional[List[str]] = None) -> None:
    """Entrypoint for image tagging."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, TAG_IMAGES_ARGS)

    kwargs = construct_kwargs(args)
    with task_context():
        tag_images(**kwargs)



import functools
import logging
import re
import yaml
from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor
import os
from typing import Any, Union, Optional, List, Dict, Tuple

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .container_image_pusher import ContainerImagePusher
from .utils.misc import (
    run_entrypoint,
    get_internal_container_repo_name,
    log_step,
    get_pyxis_ssl_paths,
    run_with_retries,
    get_basic_auth,
)
from .quay_client import QuayClient
from .utils.misc import parse_index_image, pyxis_get_repo_metadata
from .models import BuildIndexImageParam
from pubtools.tracing import get_trace_wrapper

tx = get_trace_wrapper()
LOG = logging.getLogger("pubtools.quay")


class OperatorPusher:
    """
    Add operator bundles to index images and push them to Quay.

    No validation is performed, push items are expected to be correct.
    """

    def __init__(
        self, push_items: List[Any], task_id: str, target_settings: Dict[str, Any]
    ) -> None:
        """
        Initialize.

        Args:
            push_items ([ContainerPushItem]):
                List of push items.
            task_id (str):
                task id
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.target_settings = target_settings
        self.task_id = task_id

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")
        self._version_items_mapping: Dict[str, Any] = {}
        self.ocp_versions_resolved: Dict[str, List[str]] = {}

    @staticmethod
    def _get_immutable_tag(push_item: Any) -> Union[str, Any]:
        """
        Return immutable tag from operator push item production tags.

        Args:
            push_item (ContainerPushItem):
                Operator push item.

        Returns (str):
            Immutable tag
        """
        # NOTE: while this was copied from pub, '[0]' had to be added here to work properly
        if push_item.metadata["v_r"] in list(push_item.metadata["tags"].values())[0]:
            return push_item.metadata["v_r"]
        # if v_r tag is not in destination tags, return tag with the most numbers in it
        # This code is usually trigerred when push is not initiated from ET but manually
        # by user
        tags = []
        for tag in list(push_item.metadata["tags"].values())[0]:
            tags.append((len(re.split(r"\d+", tag)), tag))

        return sorted(tags)[-1][1]

    def public_bundle_ref(self, push_item: Any) -> str:
        """
        Get public reference of a bundle image.

        It will be used by IIB to access the bundle image.

        Args:
            push_item (ContainerPushItem):
                Operator push item.

        Returns (str):
            Customer-visible bundle reference.
        """
        repository = list(push_item.metadata["tags"].keys())[0]
        # tags are the same for each destination repo, so any combination should work
        return "{0}/{1}:{2}".format(
            self.target_settings["docker_settings"]["docker_reference_registry"][0],
            repository,
            self._get_immutable_tag(push_item),
        )

    def pyxis_get_ocp_versions(self, push_item: Any) -> List[str]:
        """
        Get a list of supported ocp versions from Pyxis.

        Args:
            push_item: (ContainerPushItem)
                Push item for which the OCP version range will be found out.

        Returns ([str]):
            Supported OCP versions as returned by Pyxis.
        """
        cert, key = get_pyxis_ssl_paths(self.target_settings)

        ocp_versions = push_item.metadata["com.redhat.openshift.versions"]
        LOG.info("Getting OCP versions of '{0}' from Pyxis.".format(ocp_versions))

        args = ["--pyxis-server", self.target_settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--organization", self.target_settings["iib_organization"]]
        args += ["--ocp-versions-range", ocp_versions]
        env_vars: Dict[str, str] = {}

        data = run_entrypoint(
            (
                "pubtools-pyxis",
                "console_scripts",
                "pubtools-pyxis-get-operator-indices",
            ),
            "pubtools-pyxis-get-operator-indices",
            args,
            env_vars,
        )

        if not data:
            msg = "Pyxis has returned no OCP versions for '{0}' specified in build {1}.".format(
                ocp_versions, push_item.metadata["build"]["build_id"]
            )
            raise ValueError(msg)

        # Versions returned by Pyxis don't contain 'v' at the front (4.5 -> v4.5)
        return ["v{0}".format(item["ocp_version"]) for item in data]

    @property
    def version_items_mapping(self) -> Dict[str, List[Any]]:
        """
        Generate mapping of OCP version -> push_items.

        The mapping describes which operator bundles should be added to which index images.

        Returns ({str: [ContainerPushItem]})
            Mapping of OCP version -> Push items
        """
        if not self._version_items_mapping:
            for item in self.push_items:
                ocp_versions = item.metadata["com.redhat.openshift.versions"]
                # we haven't yet encountered this pattern, contact Pyxis for resolution

                if ocp_versions not in self.ocp_versions_resolved:
                    self.ocp_versions_resolved[ocp_versions] = self.pyxis_get_ocp_versions(item)

                for version in self.ocp_versions_resolved[ocp_versions]:
                    self._version_items_mapping.setdefault(version, []).append(item)

        return self._version_items_mapping

    def get_deprecation_list(self, version: str) -> Optional[List[str]]:
        """
        Get bundles to be deprecated in the index image.

        If deprecation list URL isn't in the target settings, None is returned.

        Args:
            version: (str)
                version for which deprecation list will be fetched.

        Returns:
            list(str)|None: list of bundles to be deprecated in the index image. or None if
                            deprecation list URL was not specified.
        """

        def _get_requests_session() -> requests.Session:
            session = requests.Session()
            retry = Retry(
                total=6,
                read=6,
                connect=6,
                backoff_factor=0.8,
                status_forcelist=(500, 502, 503, 504),
            )
            adapter = HTTPAdapter(max_retries=retry)
            session.mount("http://", adapter)
            session.mount("https://", adapter)

            return session

        if not self.target_settings.get("iib_deprecation_list_url"):
            return None

        deprecation_list = []
        deprecation_list_url = "{0}/{1}.yml/raw?ref=master".format(
            self.target_settings["iib_deprecation_list_url"].rstrip("/"),
            version.replace(".", "_"),
        )
        registry_url = self.target_settings["docker_settings"]["docker_reference_registry"][0]

        LOG.info("Getting the deprecation list for OCP version {0}".format(version))
        session = _get_requests_session()
        response = session.get(url=deprecation_list_url, timeout=10)
        if not response.ok:
            LOG.error(
                "Could not retrieve deprecation list after multiple attempts."
                " Status Code {0}".format(response.status_code)
            )
            response.raise_for_status()

        try:
            yaml_response = yaml.safe_load(response.text)
            if yaml_response:
                deprecation_list = [
                    "{0}/{1}".format(registry_url, bundle_path)
                    for pkg_deprecation_list in yaml_response.values()
                    for bundle_path in pkg_deprecation_list
                ]
        except Exception:
            LOG.error("Data in {0} is invalid".format(deprecation_list_url))
            raise

        LOG.info("Deprecation list retrieved successfully")
        return sorted(deprecation_list)

    @classmethod
    def pubtools_iib_get_common_args(
        cls, target_settings: Dict[str, Any]
    ) -> Tuple[List[str], Dict[str, str]]:
        """
        Create an argument list common for all pubtools-iib operations.

        Target settings are used to set the values of the arguments

        Args:
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.
        Returns (([str]), {str:str}):
            Tuple of arguments and environment variables to be used when calling pubtools-iib.
        """
        args = ["--iib-server", target_settings["iib_server"]]
        args += ["--iib-krb-principal", target_settings["iib_krb_principal"]]

        if target_settings.get("iib_overwrite_from_index", False):
            args += ["--overwrite-from-index"]
        if "iib_krb_ktfile" in target_settings:
            args += ["--iib-krb-ktfile", target_settings["iib_krb_ktfile"]]

        env_vars = {}
        if "iib_overwrite_from_index_token" in target_settings:
            env_vars["OVERWRITE_FROM_INDEX_TOKEN"] = target_settings[
                "iib_overwrite_from_index_token"
            ]

        if target_settings.get("check_related_images"):
            args += ["--check-related-images"]

        if target_settings.get("iib_build_timeout"):
            args += ["--build-timeout", str(target_settings["iib_build_timeout"])]

        return (args, env_vars)

    @classmethod
    @tx.instrument_func()
    def iib_add_bundles(
        cls,
        bundles: Optional[List[str]] = None,
        archs: Optional[List[str]] = None,
        index_image: Optional[str] = None,
        deprecation_list: Union[List[str], str, None] = None,
        build_tags: Optional[List[str]] = None,
        target_settings: Dict[str, Any] = {},
    ) -> Any:
        """
        Construct and execute pubtools-iib command to add bundles to index image.

        Args:
            bundles ([str]):
                External URLs to bundle images to be added to the index image.
            archs ([str]):
                Architectures to build for.
            index_image (str):
                Index image to add the bundles to.
            deprecation_list ([str]|str):
                List of bundles to be deprecated. Accepts both str (csv) and a list.
            build_tags ([str]):
                Extra tags that the new index image should be tagged with.
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.

        Returns (dict):
            Build details provided by IIB.
        """
        LOG.info(
            "Requesting IIB to add bundles '{0}' to index image '{1}'".format(bundles, index_image)
        )
        args, env_vars = cls.pubtools_iib_get_common_args(target_settings)

        if index_image:
            args += ["--index-image", index_image]
        if bundles:
            for bundle in bundles:
                args += ["--bundle", bundle]
        if archs:
            for arch in archs:
                args += ["--arch", arch]
        # inconsistent way of presenting multiple arguments...
        if deprecation_list and isinstance(deprecation_list, str):
            args += ["--deprecation-list", deprecation_list]
        elif deprecation_list and isinstance(deprecation_list, list):
            args += ["--deprecation-list", ",".join(deprecation_list)]
        if build_tags:
            for build_tag in build_tags:
                args += ["--build-tag", build_tag]

        try:
            return run_entrypoint(
                ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
                "pubtools-iib-add-bundles",
                args,
                env_vars,
            )
        except SystemExit:
            return False

    @classmethod
    @tx.instrument_func()
    def iib_remove_operators(
        cls,
        operators: Optional[List[str]] = None,
        archs: Optional[List[str]] = None,
        index_image: Optional[str] = None,
        build_tags: Optional[List[str]] = None,
        target_settings: Dict[str, Any] = {},
    ) -> Any:
        """
        Construct and execute pubtools-iib command to remove operators from index image.

        Args:
            operators ([str]):
                Operator names to be removed from the index image.
            archs ([str]):
                Architectures to build for.
            index_image (str):
                Index image to remove the operators from.
            build_tags ([str]):
                Extra tags that the new index image should be tagged with.
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.

        Returns (dict):
            Build details provided by IIB.
        """
        LOG.info(
            "Requesting IIB to remove operators '{0}' from index image '{1}'".format(
                operators, index_image
            )
        )
        args, env_vars = cls.pubtools_iib_get_common_args(target_settings)

        if index_image:
            args += ["--index-image", index_image]
        if operators:
            for operator in operators:
                args += ["--operator", operator]
        if archs:
            for arch in archs:
                args += ["--arch", arch]
        if build_tags:
            for build_tag in build_tags:
                args += ["--build-tag", build_tag]

        return run_entrypoint(
            ("pubtools-iib", "console_scripts", "pubtools-iib-remove-operators"),
            "pubtools-iib-remove-operators",
            args,
            env_vars,
        )

    @classmethod
    @tx.instrument_func()
    def iib_add_deprecations(
        cls,
        index_image: str,
        deprecation_schema: str,
        operator_package: str,
        build_tags: Optional[List[str]] = None,
        target_settings: Dict[str, Any] = {},
    ) -> Any:
        """
        Construct and execute pubtools-iib command to add deprecations of an operator package.

        Args:
            index_image (str):
                Index image to add the bundles to.
            deprecation_schema (str):
                JSON formatted deprecation schema.
            operator_package (str):
                Operator package to add deprecations to.
            build_tags ([str]):
                Extra tags that the new index image should be tagged with.
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.

        Returns (dict):
            Build details provided by IIB.
        """
        LOG.info(
            "Requesting IIB to add deprecations to package '{0}' in index image '{1}'".format(
                operator_package, index_image
            )
        )
        args, env_vars = cls.pubtools_iib_get_common_args(target_settings)

        if index_image:
            args += ["--index-image", index_image]
        if deprecation_schema:
            args += ["--deprecation-schema", deprecation_schema]
        if operator_package:
            args += ["--operator-package", operator_package]
        if build_tags:
            for build_tag in build_tags:
                args += ["--build-tag", build_tag]

        return run_entrypoint(
            ("pubtools-iib", "console_scripts", "pubtools-iib-add-deprecations"),
            "pubtools-iib-add-deprecations",
            args,
            env_vars,
        )

    def get_existing_index_images(self, quay_client: QuayClient) -> List[Tuple[str, str, str]]:
        """
        Return existing index images for non-fbc push items.

        Args:
            quay_client (QuayClient): quay_client_instance

        Returns [(digest, tag, repo)]:
            List of tuples containing digest and tag of existing index image
        """
        image_schema = "{host}/{namespace}/{repo}"
        iib_repo = get_internal_container_repo_name(
            self.target_settings["quay_operator_repository"]
        )
        index_image_repo = image_schema.format(
            host=self.quay_host,
            namespace=self.target_settings.get(
                "quay_operator_namespace", self.target_settings["quay_namespace"]
            ),
            repo=iib_repo,
        )
        current_index_images = []

        manifest_list: Any = {}
        for version, items in sorted(self.version_items_mapping.items()):
            items_opted_in, failed_items = self._get_fbc_opted_in_items()
            if not self._get_non_fbc_items_for_version(items, version, items_opted_in):
                continue

            image_ref = "{0}:{1}".format(index_image_repo, version)
            try:
                manifest_list = quay_client.get_manifest(
                    image_ref, media_type=QuayClient.MANIFEST_LIST_TYPE
                )
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404 or e.response.status_code == 401:
                    continue
                else:
                    raise
            for manifest in manifest_list["manifests"]:
                current_index_images.append(
                    (
                        manifest["digest"],
                        version,
                        self.target_settings["quay_operator_repository"],
                    )
                )
        return list(set(current_index_images))

    @log_step("Verify bundles presence")
    def ensure_bundles_present(self) -> bool:
        """
        Make sure bundles are present in Quay.

        Wait until pushed images become present in Quay,
        and return False if too much time has passed.
        """
        bundles = [self.public_bundle_ref(i) for i in self.push_items]
        for bundle in bundles:
            registry = bundle.split("/", 1)[0]
            username, password = get_basic_auth(registry) or ("", "")
            quay_client = QuayClient(username, password, registry)
            try:
                get_manifest_partial = functools.partial(quay_client.get_manifest, bundle)
                run_with_retries(
                    get_manifest_partial,
                    "Verify bundle presence",
                    self.target_settings.get("verify_bundle_tries", 5),
                    self.target_settings.get("verify_bundle_wait_time_increase", 20),
                )
            except Exception:
                LOG.error("Bundle {0} cannot be reached".format(bundle))
                return False
            LOG.info("Bundle {0} is present".format(bundle))
        return True

    def _get_fbc_opted_in_items(self) -> Tuple[Dict[int, bool], Dict[int, bool]]:
        """Get items that are opted in for fbc.

        An item needs to be targeted for repos with fbc_opt_in set to True and
        ocp versions needs to be higher than 4.10. Inconsistencies in versions
        (like support for both > 4.10 and <= 4.10) results  in item error.
        """
        repos_opted_in = {}
        items_opted_in = {}
        failed_items = {}

        # We need to load pyxis resolved versions at this point
        versions_mapping = self.version_items_mapping  # noqa: F841

        for item in self.push_items:
            for repo in item.metadata["tags"]:
                if repo not in repos_opted_in:
                    repos_opted_in[repo] = pyxis_get_repo_metadata(repo, self.target_settings).get(
                        "fbc_opt_in", False
                    )
            item_fbc_opt_in = [repos_opted_in[repo] for repo in item.metadata["tags"]]
            if not (all(item_fbc_opt_in) or not any(item_fbc_opt_in)):
                failed_items[id(item)] = True
                item.add_error("NOTPUSHED", "Inconsistent fbc opt in")
                LOG.error("Inconsistent fbc opt in for item {i}".format(i=item))
                continue
            elif all(item_fbc_opt_in):
                items_opted_in[id(item)] = True
            else:
                items_opted_in[id(item)] = False

            ocp_versions = item.metadata["com.redhat.openshift.versions"]
            if (
                [
                    version
                    for version in self.ocp_versions_resolved[ocp_versions]
                    if tuple([int(x) for x in version.replace("v", "").split(".")]) < (4, 11)
                ]
                and [
                    version
                    for version in self.ocp_versions_resolved[ocp_versions]
                    if tuple([int(x) for x in version.replace("v", "").split(".")]) > (4, 10)
                ]
                and items_opted_in[id(item)]
            ):
                item.add_error(
                    "INVALIDFILE",
                    "Cannot push item to index image "
                    "as it supports both <= 4.10 and >= 4.11 and is opted in FBC: {item}".format(
                        item=item
                    ),
                )
                LOG.error(
                    "Cannot push item to index image "
                    "as it supports both <= 4.10 and >= 4.11 and is opted in FBC: {item}".format(
                        item=item
                    )
                )
                failed_items[id(item)] = True
        return items_opted_in, failed_items

    def _get_non_fbc_items_for_version(
        self, items: List[Any], version: str, items_opted_in: Dict[int, bool]
    ) -> List[Any]:
        """Return non fbc items for given ocp version.

        Args:
            items: List[ContainerPushItem]
                list of push items
            version: str
                ocp version for which items should be returned.
            items_opted_in: Dict[int, bool]
                list of push items opted in fbc.
        Returns List[ContainerPushItem]:
            list of items not opted in fbc
        """
        non_fbc_items = []
        osev_tuple = tuple([int(x) for x in version.replace("v", "").split(".")])
        for item in items:
            if not items_opted_in[id(item)] or (items_opted_in[id(item)] and osev_tuple <= (4, 10)):
                non_fbc_items.append(item)
            elif items_opted_in[id(item)] and osev_tuple >= (4, 11):
                LOG.warning(
                    "Skipping {i}".format(i=item)
                    + "from iib build as it's opted in for FBC and targeting OCP version >=4.11"
                )
        return non_fbc_items

    def _create_item_groups_for_version(
        self,
        non_fbc_items: List[Any],
        version: str,
        is_hotfix: bool = False,
        is_prerelease: bool = False,
    ) -> Dict[str, Dict[str, Any]]:
        """Iterate thought non fbc items and group those together based on destination tag.

        Args:
            non_fbc_items: List[ContainerPushItem]
                list of items not opted in fbc
            version: str
                ocp version
            is_hotfix: bool
                flag indicating items are for hotfix push
            is_prerelease: bool
                flag indicating items are for prerelease push
        Returns Dict[str, Dict[str, Any]]:
            Dictionary of items grouped by ocp version
        """
        item_groups: Dict[str, Dict[str, Any]] = {
            version: {
                "items": [],
                "overwrite": True,
                "destination_tags": [version],
            }
        }
        if is_hotfix or is_prerelease:
            for item in non_fbc_items:
                if is_hotfix:
                    tag_part = item.metadata["com.redhat.hotfix"]
                    origin = item.origin
                else:
                    tag_part = "prerelease-{0}".format(
                        item.metadata["operators.operatorframework.io.bundle.package.v1"]
                    )
                    origin = (
                        item.origin
                        + "-"
                        + item.metadata["operators.operatorframework.io.bundle.package.v1"]
                    )
                dst_tag = "{0}-{1}-{2}".format(
                    version,
                    tag_part,
                    item.origin.split("-")[1].replace(":", "-"),
                )
                item_groups.setdefault(
                    origin,
                    {
                        "items": [],
                        "overwrite": False,
                        "destination_tags": [],
                    },
                )
                item_groups[origin]["items"].append(item)
                item_groups[origin]["destination_tags"].append(dst_tag)
                if is_prerelease:
                    floating_dst_tag = "{0}-{1}".format(version, tag_part)
                    item_groups[origin]["destination_tags"].append(floating_dst_tag)
        else:
            for item in non_fbc_items:
                item_groups[version]["items"].append(item)
        return item_groups

    @log_step("Build index images")
    def build_index_images(self) -> Dict[str, Dict[Any, Any]]:
        """
        Perform the 'build' part of the operator workflow.

        This workflow is a part of push-docker operation.
        The workflow can be summarized as:
        - Use Pyxis to parse 'com.redhat.openshift.versions'
        - Filter out push items which opted in to FBC and shouldn't be pushed
        - Set extra attributes or push items are for prerelease or for hotfix
        - Get deprecation list for a given version (list of bundles to be deprecated)
        - Create mapping of which bundles should be pushed to which index image versions
        - Contact IIB to add the bundles to the index images with multiple threads

        Returns ({str:dict}):
            Dictionary containing IIB results and signing keys for all OPM versions. Data will be
            used in operator signing. Dictionary structure:
            {
            <target_tag>: {
            "iib_result": (...) (object returned by iiblib)
            "signing_keys": [...] (list of signing keys to be used for signing)
            "destination_tags": [...] (list of destination tags)
            }
            }
        """
        iib_results = {}

        items_opted_in, failed_items = self._get_fbc_opted_in_items()

        # if any of push items failed due to fbc issues, return early and skip all iib operations
        if failed_items:
            return {}

        build_index_image_params = []

        for version, items in sorted(self.version_items_mapping.items()):
            non_fbc_items = self._get_non_fbc_items_for_version(items, version, items_opted_in)

            is_hotfix = any([item.metadata.get("com.redhat.hotfix") for item in non_fbc_items])
            is_prerelease = any(
                [item.metadata.get("com.redhat.prerelease") for item in non_fbc_items]
            )
            is_advisory_source = all(
                [re.match(r"^[A-Z0-9:\-]{4,40}$", item.origin) for item in non_fbc_items]
            )
            if is_hotfix and not is_advisory_source:
                raise ValueError("Cannot push hotfixes without an advisory")
            if is_prerelease and not is_advisory_source:
                raise ValueError("Cannot push pre release without an advisory")

            item_groups = self._create_item_groups_for_version(
                non_fbc_items, version, is_hotfix, is_prerelease
            )

            # Get deprecation list
            deprecation_list = self.get_deprecation_list(version)
            for group, group_info in item_groups.items():
                if not group_info["items"]:
                    continue
                tag = version
                index_image = "{image_repo}:{tag}".format(
                    image_repo=self.target_settings["iib_index_image"], tag=tag
                )
                build_tags = []
                if is_prerelease:
                    build_tags.append(
                        "{0}-{1}-{2}".format(
                            index_image.split(":")[1], self.task_id, group.replace(":", "-")
                        )
                    )
                else:
                    build_tags.append("{0}-{1}".format(index_image.split(":")[1], self.task_id))

                bundles = [self.public_bundle_ref(i) for i in group_info["items"]]
                signing_keys = sorted(
                    list(set([item.claims_signing_key for item in group_info["items"]]))
                )

                target_settings = self.target_settings.copy()
                if not group_info["overwrite"]:
                    target_settings["iib_overwrite_from_index"] = False
                    target_settings["iib_overwrite_from_index_token"] = ""
                else:
                    target_settings = self.target_settings

                # save parameters which are used to build index image in IIB.
                build_index_image_params.append(
                    BuildIndexImageParam(
                        bundles=bundles,
                        index_image=index_image,
                        deprecation_list=deprecation_list,
                        build_tags=build_tags,
                        target_settings=target_settings,
                        tag=tag,
                        signing_keys=signing_keys,
                        destination_tags=group_info["destination_tags"],
                        origin=group,
                    )
                )

        num_thread_build_index_images = self.target_settings.get("num_thread_build_index_images", 5)

        orig_environ = os.environ.copy()
        _, environ_vars = self.pubtools_iib_get_common_args(self.target_settings)
        for key in environ_vars:
            os.environ[key] = environ_vars[key]

        try:
            with ThreadPoolExecutor(max_workers=num_thread_build_index_images) as executor:
                future_results = {
                    executor.submit(
                        lambda param: self.iib_add_bundles(
                            bundles=param.bundles,
                            index_image=param.index_image,
                            deprecation_list=param.deprecation_list,
                            build_tags=param.build_tags,
                            target_settings=param.target_settings,
                        ),
                        param,
                    ): param
                    for param in build_index_image_params
                }
                for future in futures.as_completed(future_results):
                    build_details = future.result()
                    param = future_results[future]
                    iib_results[param.tag + "-" + param.origin] = {
                        "iib_result": build_details,
                        "signing_keys": param.signing_keys,
                        "destination_tags": param.destination_tags,
                    }
        finally:
            os.environ.update(orig_environ)
            to_delete = [key for key in os.environ if key not in orig_environ]
            for key in to_delete:
                del os.environ[key]

        return iib_results

    @log_step("Push index images to Quay")
    def push_index_images(
        self, iib_results: Dict[Any, Any], tag_suffix: Optional[str] = None
    ) -> None:
        """
        Push index images which were built in the previous stage to Quay.

        Args:
            iib_results (dict):
                IIB results returned by the build stage
            tag_suffix (str):
                extra tag suffix applied to iib version tags if specified
        """
        image_schema = "{host}/{namespace}/{repo}"
        image_schema_tag = "{host}/{namespace}/{repo}:{tag}"
        index_image_repo = image_schema.format(
            host=self.quay_host,
            namespace=self.target_settings.get(
                "quay_operator_namespace", self.target_settings["quay_namespace"]
            ),
            repo=get_internal_container_repo_name(self.target_settings["quay_operator_repository"]),
        )

        for version, results in iib_results.items():
            build_details = results.get("iib_result", None)
            if not build_details:
                continue

            _, tag = build_details.index_image.split(":", 1)
            iib_feed, iib_namespace, iib_intermediate_repo = parse_index_image(build_details)
            permanent_index_image = image_schema_tag.format(
                host=iib_feed,
                namespace=iib_namespace,
                repo=iib_intermediate_repo,
                tag=build_details.build_tags[0],
            )
            dest_images = [
                "{0}:{1}".format(index_image_repo, dst_tag)
                for dst_tag in results["destination_tags"]
            ]

            # We don't use permanent index image here because we always want to overwrite
            # production tags with the latest index image (in case of parallel pushes)
            index_image_ts = self.target_settings.copy()
            index_image_ts["dest_quay_user"] = index_image_ts.get(
                "index_image_quay_user", index_image_ts["dest_quay_user"]
            )
            index_image_ts["dest_quay_password"] = index_image_ts.get(
                "index_image_quay_password", index_image_ts["dest_quay_password"]
            )

            ContainerImagePusher.run_tag_images(
                build_details.index_image, dest_images, True, index_image_ts
            )
            if tag_suffix:
                _dest_images = []
                for _dest_tag in results["destination_tags"]:
                    _dest_images.append(
                        "{0}:{1}-{2}".format(index_image_repo, _dest_tag, tag_suffix)
                    )
                ContainerImagePusher.run_tag_images(
                    permanent_index_image, _dest_images, True, index_image_ts
                )



from typing import Any, Union

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


# pylint: disable=bad-option-value,useless-object-inheritance
class QuaySession(object):
    """Helper class to support Quay requests and authentication."""

    def __init__(
        self,
        hostname: Union[str, None] = None,
        retries: int = 3,
        backoff_factor: int = 2,
        verify: bool = False,
        api: str = "docker",
    ) -> None:
        """
        Initialize.

        Args:
            hostname (str)
                hostname of Quay service.
            retries (int)
                number of http retries.
            backoff_factor (int)
                backoff factor to apply between attempts after the second try.
            verify (bool)
                enable/disable SSL CA verification.
            api (str):
                Which API queries to construct. Supported values: 'docker', 'quay'
        """
        if api not in ("docker", "quay"):
            raise ValueError("Unknown API type: '{0}'".format(api))
        self.api = api

        self.session = requests.Session()
        self.hostname = hostname or "quay.io"
        self.session.verify = verify
        self.session.headers["Host"] = self.hostname

        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=set(range(500, 512)),
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

    def get(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        HTTP GET request against Quay server API.

        Args:
            endpoint (str):
                Endpoint of the request.
            kwargs:
                Additional arguments to add to the requests method.
        Returns:
            requests.Response: A response object.
        """
        if "timeout" not in kwargs:
            kwargs["timeout"] = 10
        return self.session.get(self._api_url(endpoint), **kwargs)

    def post(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        HTTP POST request against Quay server API.

        Args:
            endpoint (str):
                Endpoint of the request.
            kwargs:
                Additional arguments to add to the requests method.
        Returns:
            requests.Response: A response object.
        """
        if "timeout" not in kwargs:
            kwargs["timeout"] = 10
        return self.session.post(self._api_url(endpoint), **kwargs)

    def put(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        HTTP PUT request against Quay server API.

        Args:
            endpoint (str):
                Endpoint of the request.
            kwargs:
                Additional arguments to add to the requests method.
        Returns:
            requests.Response: A response object.
        """
        if "timeout" not in kwargs:
            kwargs["timeout"] = 10
        return self.session.put(self._api_url(endpoint), **kwargs)

    def delete(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        HTTP DELETE request against Quay server API.

        Args:
            endpoint (str):
                Endpoint of the request.
            kwargs:
                Additional arguments to add to the requests method.
        Returns:
            requests.Response: A response object.
        """
        if "timeout" not in kwargs:
            kwargs["timeout"] = 10
        return self.session.delete(self._api_url(endpoint), **kwargs)

    def request(self, method: str, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        HTTP generic request against Quay server API.

        Args:
            method (str):
                REST API method of the request (GET, POST, PUT, DELETE).
            endpoint (str):
                Endpoint of the request.
            kwargs:
                Additional arguments to add to the requests method.
        Returns:
            requests.Response: A response object.
        """
        if "timeout" not in kwargs:
            kwargs["timeout"] = 10
        return self.session.request(method, self._api_url(endpoint), **kwargs)

    def _api_url(self, endpoint: str) -> str:
        """
        Generate full url of the API endpoint.

        Args:
            endpoint (str)
                API specific endpoint for the request.
        Returns:
            str: Full URL of the endpoint.
        """
        if self.api == "docker":
            schema = "{0}{1}/v2/{2}"
        elif self.api == "quay":
            schema = "{0}{1}/api/v1/{2}"

        if "http://" not in self.hostname and "https://" not in self.hostname:
            return schema.format("https://", self.hostname.rstrip("/"), endpoint)
        else:
            return schema.format("", self.hostname.rstrip("/"), endpoint)

    def set_auth_token(self, token: str) -> None:
        """
        Set a Bearer auth token for the authentication.

        Args:
            token (str):
                Bearer token.
        """
        self.session.headers["Authorization"] = "Bearer {0}".format(token)



import mock

import requests_mock

from pubtools._quay.signer_wrapper import CosignSignerWrapper


def test_remove_signatures(cosign_signer_settings, fake_cert_key_paths, dest_manifest_list):
    with mock.patch("pubtools._quay.signer_wrapper.run_entrypoint") as mock_run_entry_point:
        with requests_mock.Mocker() as m:
            m.get(
                "https://test-quay.io/v2/testing/testing----repository/manifests/sha256:123456789",
                json=dest_manifest_list,
                headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
            )
            mock_run_entry_point.return_value = (
                True,
                ["quay.io/testing/repository:sha256-123456789.sig"],
            )
            sw = CosignSignerWrapper(
                config_file="fake-config-file", settings=cosign_signer_settings
            )
            # TODO: uncomment when cosign removing signatures is enabled
            # with mock.patch(
            #     "pubtools._quay.signer_wrapper.QuayApiClient.delete_tag"
            # ) as mock_delete_tag:
            sw.remove_signatures([("sha256:123456789", "tag", "testing/repository")])
            # mock_delete_tag.assert_not_called()
            # TODO: uncomment when removing signatures is enabled
            # mock_delete_tag.assert_called_once_with(
            #    "testing/testing----repository", "sha256-123456789.sig"
            # )


def test_remove_signatures_failure(cosign_signer_settings, fake_cert_key_paths, dest_manifest_list):
    with mock.patch("pubtools._quay.signer_wrapper.run_entrypoint") as mock_run_entry_point:
        with requests_mock.Mocker() as m:
            m.get(
                "https://test-quay.io/v2/testing/testing----repository/manifests/sha256:123456789",
                json=dest_manifest_list,
                headers={"Content-Type": "application/vnd.docker.distribution.manifest.v2+json"},
            )
            mock_run_entry_point.return_value = (
                False,
                "test-error",
            )
            sw = CosignSignerWrapper(
                config_file="fake-config-file", settings=cosign_signer_settings
            )
            # TODO: uncomment when cosign removing signatures is enabled
            # with mock.patch(
            #     "pubtools._quay.signer_wrapper.QuayApiClient.delete_tag"
            # ) as mock_delete_tag:
            sw.remove_signatures([("sha256:123456789", "tag", "testing/repository")])
            # mock_delete_tag.assert_not_called()



import os
import tempfile
import uuid
import subprocess
import logging
import json
import base64
from dataclasses import dataclass
from concurrent import futures
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, cast, List, Dict, Union, Optional, Set

from .quay_client import QuayClient
from .utils.misc import get_internal_container_repo_name, log_step, retry
from .command_executor import LocalExecutor
from .quay_api_client import QuayApiClient
from .exceptions import ManifestTypeError
from .types import ManifestList

LOG = logging.getLogger("pubtools.quay")


@dataclass
class DigestSecurityManifest:
    """Class for keeping track of which digest is associated with which security manifest."""

    digest: str
    security_manifest_path: str


class SecurityManifestPusher:
    """Class that pushes container security manifests."""

    COSIGN_TRIANGULATE_TYPES = ("attestation", "sbom", "signature")

    def __init__(self, push_items: List[object], target_settings: Dict[str, Any]) -> None:
        """
        Initialize.

        Args:
            push_items ([ContainerPushItem]):
                List of push items.
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.target_settings = target_settings
        # TODO: change this to whatever is decided
        self.cosign_private_key_path = self.target_settings["cosign_private_key_path"]
        self.cosign_public_key_path = self.target_settings["cosign_public_key_path"]
        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")

        self._src_quay_client: Optional[QuayClient] = None
        self._dest_quay_client: Optional[QuayClient] = None
        self._dest_quay_api_client: Optional[QuayApiClient] = None

    @property
    def src_quay_client(self) -> QuayClient:
        """Create and access QuayClient for source image."""
        if self._src_quay_client is None:
            self._src_quay_client = QuayClient(
                self.target_settings["source_quay_user"],
                self.target_settings["source_quay_password"],
                self.target_settings.get("source_quay_host") or self.quay_host,
            )
        return self._src_quay_client

    @property
    def dest_quay_client(self) -> QuayClient:
        """Create and access QuayClient for dest image."""
        if self._dest_quay_client is None:
            self._dest_quay_client = QuayClient(
                self.target_settings["dest_quay_user"],
                self.target_settings["dest_quay_password"],
                self.quay_host,
            )
        return self._dest_quay_client

    @property
    def dest_quay_api_client(self) -> QuayApiClient:
        """Create and access QuayApiClient for dest image."""
        if self._dest_quay_api_client is None:
            self._dest_quay_api_client = QuayApiClient(
                self.target_settings["dest_quay_api_token"], self.quay_host
            )
        return self._dest_quay_api_client

    @classmethod
    @retry("Get security manifest")
    def cosign_get_security_manifest(self, image_ref: str, output_file: str) -> bool:
        """
        Use cosign to get security manifest from an image and save it to a file.

        Args:
            image_ref (str):
                Image reference to get the manifest from.
            output_file (str):
                File where to save the manifest.
        Returns (bool):
            Whether the manifest was gathered successfully (True) or not (False).
        """
        cmd = ["cosign", "download", "sbom", image_ref, "--output-file", output_file]
        LOG.info(f"Running command '{' '.join(cmd)}'")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        if result.returncode and "no sbom attached to reference" in result.stdout:
            LOG.warning(f"Command {' '.join(cmd)} has failed: {result.stdout}")
            return False
        # If this string was not matched, the error is unexpected - raise
        elif result.returncode:
            raise RuntimeError(f"Command {' '.join(cmd)} has failed with an error: {result.stdout}")

        return True

    @retry("Get existing attestation")
    def cosign_get_existing_attestation(
        self,
        image_ref: str,
        output_file: str,
        rekor_url: Optional[str] = None,
        skip_verify_rekor: bool = False,
    ) -> bool:
        """
        Use cosign to verify and get an attestation, if it exists.

        Args:
            image_ref (str):
                Image reference to get the manifest from.
            output_file (str):
                File where to save the attestation.
            rekor_url (str):
                URL of the rekor instance to use. If unset, default will be used.
            skip_verify_rekor (bool):
                Whether to skip rekor log verification. This option is recommended for when an
                attestation was created without uploading a transparency log to rekor. rekor_url
                parameter is ignored if this option is enabled.
        Returns (bool):
            Whether the attestation was gathered successfully (True) or not (False).
        """
        cmd = [
            "cosign",
            "verify-attestation",
            "--key",
            self.cosign_public_key_path,
            image_ref,
            "--output-file",
            output_file,
        ]
        if skip_verify_rekor:
            cmd.insert(2, "--insecure-ignore-tlog=true")
        elif rekor_url:
            cmd.insert(2, f"--rekor-url={rekor_url}")
        LOG.info(f"Running command '{' '.join(cmd)}'")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        # If no errors are displayed, it means that the attestation doesn't exist
        if result.returncode and "no matching attestations: \n" in result.stdout:
            LOG.warning(f"Command {' '.join(cmd)} has failed: {result.stdout}")
            return False
        # if an unexpected error is displayed, raise an error
        elif result.returncode:
            raise RuntimeError(f"Command {' '.join(cmd)} has failed with an error: {result.stdout}")

        return True

    @retry("Attest security manifest")
    def cosign_attest_security_manifest(
        self,
        security_manifest_path: str,
        image_ref: str,
        rekor_url: Optional[str] = None,
        skip_upload_rekor: bool = False,
    ) -> None:
        """
        Use cosign to attest a security manifest and push the created image to the destination.

        Args:
            security_manifest_path (str):
                Path to the security manifest.
            image_ref (str):
                Image to which the security manifest should be attested to.
            rekor_url (str):
                URL of the rekor instance to use. If unset, default will be used.
            skip_upload_rekor (bool):
                Whether to skip uploading transparency log to rekor. rekor_url parameter is ignored
                if this option is enabled.
        Raises:
            RuntimeError:
                If the command fails.
        """
        cmd = [
            "cosign",
            "attest",
            "--predicate",
            security_manifest_path,
            "--key",
            self.cosign_private_key_path,
            "-y",
            image_ref,
        ]
        if skip_upload_rekor:
            cmd.insert(2, "--tlog-upload=false")
        elif rekor_url:
            cmd.insert(2, f"--rekor-url={rekor_url}")
        LOG.info(f"Running command '{' '.join(cmd)}'")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        if result.returncode:
            LOG.error(f"Command {' '.join(cmd)} has failed: {result.stdout}")
            raise RuntimeError(f"Creating attestation to image {image_ref} has failed")

    @classmethod
    def cosign_triangulate_image(
        self, image_ref: str, dir_path: str, image_type: str = "attestation"
    ) -> str:
        """
        Use cosign to get the reference to the attestation image.

        Args:
            image_ref (str):
                Image whose attestation image reference to get.
            dir_path (str):
                Path where a file containing the reference should be created.
            image_type (str):
                Type of image being triangulated.
        Returns (str):
            Reference of the attestation image.
        Raises:
            RuntimeError:
                If the command fails.
        """
        if image_type not in self.COSIGN_TRIANGULATE_TYPES:
            raise ValueError(
                f"Image type '{image_type}' needs to be one of {self.COSIGN_TRIANGULATE_TYPES}"
            )

        reference_file = os.path.join(dir_path, f"{image_type}_reference_{uuid.uuid4().hex}.json")
        cmd = [
            "cosign",
            "triangulate",
            f"--type={image_type}",
            image_ref,
            "--output-file",
            reference_file,
        ]
        LOG.info(f"Running command '{' '.join(cmd)}'")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)

        if result.returncode:
            LOG.error(f"Command {' '.join(cmd)} has failed: {result.stdout}")
            raise RuntimeError(f"Triangulating {image_type} image to image {image_ref} has failed")

        with open(reference_file, "r") as f:
            return f.read().strip()

    def get_security_manifest_from_attestation(self, file_path: str) -> Union[Dict[Any, Any], Any]:
        """
        Parse image attestation and extract a security manifest.

        Args:
            file_path (str):
                Path of a file containing the attestation.
        Returns (dict):
            Parsed attestation.
        """
        with open(file_path, "r") as f:
            attestation = json.load(f)
        data = json.loads(base64.b64decode(attestation["payload"]))["predicate"]["Data"]

        return json.loads(data)

    def security_manifest_get_products(self, security_manifest: Dict[Any, Any]) -> Set[str]:
        """
        Get a list of products from an already uploaded security manifest.

        Products are expected to be in format:
        "properties":[
        {
        "name": "product",
        "value": "product-1"
        },
        {
        "name": "product",
        "value": "product-2"
        }
        ]

        Args:
            security_manifest (dict):
                Security manifest extracted from an attestation.
        Returns (set[str]):
            List of products.
        """
        products = set()
        for property in security_manifest.get("properties", []):
            if property["name"] == "product":
                products.add(property["value"])

        return products

    def get_destination_repos(self, item: Any) -> List[str]:
        """
        Get a list of destination refs (without tag/digest) of a push item.

        Args:
            item (ContainerPushItem):
                Push item.
        Returns ([str]):
            List of destination repositories (refs without tag/digest).
        """
        dest_repos = []
        image_schema = "{host}/{namespace}/{repo}"

        for repo in sorted(item.metadata["tags"].keys()):
            internal_repo = get_internal_container_repo_name(repo)
            dest_repo = image_schema.format(
                host=self.quay_host,
                namespace=self.target_settings["quay_namespace"],
                repo=internal_repo,
            )
            dest_repos.append(dest_repo)

        return list(set(dest_repos))

    def security_manifest_remove_incompleteness_reasons(self, security_manifest_path: str) -> str:
        """
        Remove the field "incompleteness_reasons" from the security manifest.

        The field is for internal use only, and isn't a part of the CycloneDX spec.

        Args:
            security_manifest_path (str):
                Path to the extracted security manifest.

        Returns (str):
            Path to a file containing the modified security manifest.
        """
        with open(security_manifest_path, "r") as f1:
            security_manifest = json.load(f1)

        if "incompleteness_reasons" in security_manifest:
            del security_manifest["incompleteness_reasons"]

        modified_security_manifest_path = os.path.join(
            os.path.dirname(security_manifest_path),
            f"sanitized_security_manifest_{uuid.uuid4().hex}.json",
        )
        with open(modified_security_manifest_path, "w") as f2:
            json.dump(security_manifest, f2, indent=4)

        return modified_security_manifest_path

    def security_manifest_add_products(
        self, security_manifest_path: str, products: Set[str]
    ) -> str:
        """
        Add product names of the shipped image to the security manifest.

        Modified manifest will be saved to a new file in the same directory.

        Args:
            security_manifest_path (str):
                Path to the extracted security manifest.
            products (set[str]):
                Products to add to the security manifest.
        Returns (str):
            Path to a file containing the modified security manifest.
        """
        with open(security_manifest_path, "r") as f1:
            security_manifest = json.load(f1)

        security_manifest["properties"] = []
        for product in products:
            security_manifest["properties"].append({"name": "product", "value": product})

        modified_security_manifest_path = os.path.join(
            os.path.dirname(security_manifest_path),
            f"full_security_manifest_{uuid.uuid4().hex}.json",
        )
        with open(modified_security_manifest_path, "w") as f2:
            json.dump(security_manifest, f2, indent=4)

        return modified_security_manifest_path

    def delete_existing_attestation(self, image_ref: str, dir_path: str) -> None:
        """
        Delete an existing attestation image.

        This may be necessary if a new product name is to be added to an existing security manifest.
        By default, 'cosign attest' appends the new security manifest to the existing attestation,
        making it offer two manifests instead of one. It is also unable to replace an existing
        attestation by itself. So the attestation image needs to be removed manually before being
        reuploaded with the complete list of products.

        Args:
            image_ref (str):
                Image whose attestation image should be removed.
            dir_path (str):
                Path to a directory where temporary files may be created.
        """
        LOG.info(f"Removing attestation image of image {image_ref}")
        attestation_image_ref = self.cosign_triangulate_image(image_ref, dir_path)

        tag = attestation_image_ref.split(":")[-1]
        repo_path = attestation_image_ref.split(":")[0]
        # Take only last two parts of the url, which are namespace + repo
        repository = "/".join(repo_path.split("/")[-2:])
        self.dest_quay_api_client.delete_tag(repository, tag)

    def extract_security_manifest_from_attestation(self, attestation_path: str) -> str:
        """
        Parse security manifest from an attestation and save it to a file.

        Args:
            attestation_path (str):
                Path to an attestation.

        Returns (str):
            Path to the file containing the security manifest.
        """
        security_manifest = self.get_security_manifest_from_attestation(attestation_path)
        extracted_security_manifest_path = os.path.join(
            os.path.dirname(attestation_path),
            f"extracted_security_manifest_{uuid.uuid4().hex}.json",
        )
        with open(extracted_security_manifest_path, "w") as f:
            json.dump(security_manifest, f, indent=4)

        return extracted_security_manifest_path

    def merge_and_push_security_manifest(
        self,
        item: Any,
        image_manifest: DigestSecurityManifest,
        destination_repos: List[str],
        dir_path: str,
    ) -> None:
        """
        Add products to security manifest and create an attestation.

        It's possible that a push item will not contain a product name. In this case, it will not
        be included in the security manifest. This may happen with the PushDocker method.
        Args:
            item (ContainerPushItem):
                Push item.
            image_manifest (DigestSecurityManifest):
                Image manifest that will be attested.
            destination_repos ([str]):
                Destination paths (without a tag) where the attestation will be pushed.
            dir_path (str):
        """
        if item.metadata.get("product_name"):
            products = set([item.metadata["product_name"]])
        else:
            products = set()

        for repo in destination_repos:
            image_ref = f"{repo}@{image_manifest.digest}"
            attestation_file = os.path.join(dir_path, f"attestation_{uuid.uuid4().hex}.json")
            attestation_exist = self.cosign_get_existing_attestation(
                image_ref,
                attestation_file,
                self.target_settings.get("cosign_rekor_url", None),
                self.target_settings.get("cosign_sbom_skip_verify_rekor", False),
            )

            if attestation_exist:
                LOG.info(f"Existing attestation file found for image {image_ref}")
                existing_security_manifest = self.get_security_manifest_from_attestation(
                    attestation_file
                )
                existing_products = self.security_manifest_get_products(existing_security_manifest)

                # It is assumed that images with the same digest must have the same
                # content of security manifests
                if len(products - existing_products) == 0:
                    LOG.info(
                        "No new product will be added to the existing security manifest. Skipping "
                        f"the security manifest upload for image {image_ref}."
                    )
                    continue
                LOG.info(
                    f"Image {image_ref} already has an attestation. A new attestation will "
                    "be uploaded containing all product names."
                )

                self.delete_existing_attestation(image_ref, dir_path)
                products = products | existing_products

            sanitized_security_manifest_path = self.security_manifest_remove_incompleteness_reasons(
                image_manifest.security_manifest_path
            )
            if products:
                full_security_manifest_path = self.security_manifest_add_products(
                    sanitized_security_manifest_path, products
                )
            else:
                LOG.warning(
                    f"Push item {item} doesn't contain a product name. A new attestation "
                    "will be created without this information."
                )
                full_security_manifest_path = sanitized_security_manifest_path

            self.cosign_attest_security_manifest(
                full_security_manifest_path,
                image_ref,
                self.target_settings.get("cosign_rekor_url", None),
                self.target_settings.get("cosign_sbom_skip_upload_rekor", False),
            )

    def push_manifest_list_security_manifests(self, item: Any, dir_path: str) -> None:
        """
        Push all arch attestations to the manifest list digests of all destinations.

        This ensures that getting an attestation of an images specified via tag returns a result.
        The result will be all arch attestations in a JSONL format.

        Since manifest list merging workflow is enabled, the multiarch destination tags may have
        various digests and archs. Inspect each destination tag to determine which arch attestations
        are needed and what is the ML digest.

        Args:
            item (PushItem):
                Push Item to proccess
            dir_path (str):
                Directory path where artifacts will be created.
        """
        repo_schema = "{host}/{namespace}/{repo}"
        for repo, tags in sorted(item.metadata["tags"].items()):
            internal_repo = get_internal_container_repo_name(repo)
            dest_repo = repo_schema.format(
                host=self.quay_host,
                namespace=self.target_settings["quay_namespace"],
                repo=internal_repo,
            )

            for tag in tags:
                dest_ref = f"{dest_repo}:{tag}"
                LOG.info(f"Creating multiarch attestation for {dest_ref}")
                dest_digest = self.dest_quay_client.get_manifest_digest(
                    dest_ref, self.dest_quay_client.MANIFEST_LIST_TYPE
                )
                dest_digest_ref = f"{dest_repo}@{dest_digest}"
                # If the image somehow doesn't contain a ML, an error is raised
                dest_ml = cast(
                    ManifestList,
                    self.dest_quay_client.get_manifest(
                        dest_ref, media_type=self.dest_quay_client.MANIFEST_LIST_TYPE
                    ),
                )

                tag_attestations = []
                # All arch attestations are already created, gather them
                for manifest in dest_ml["manifests"]:
                    arch_ref = f"{dest_repo}@{manifest['digest']}"
                    attestation_file = os.path.join(
                        dir_path, f"attestation_{uuid.uuid4().hex}.json"
                    )
                    arch_attestation_exist = self.cosign_get_existing_attestation(
                        arch_ref,
                        attestation_file,
                        self.target_settings.get("cosign_rekor_url", None),
                        self.target_settings.get("cosign_sbom_skip_verify_rekor", False),
                    )
                    if not arch_attestation_exist:
                        LOG.warning(
                            f"Arch image {arch_ref} that is a part of {dest_ref} "
                            "doesn't have an attestation"
                        )
                        continue

                    security_manifest_path = self.extract_security_manifest_from_attestation(
                        attestation_file
                    )
                    tag_attestations.append(security_manifest_path)

                attestation_file = os.path.join(dir_path, f"attestation_{uuid.uuid4().hex}.json")
                ml_attestation_exist = self.cosign_get_existing_attestation(
                    dest_digest_ref,
                    attestation_file,
                    self.target_settings.get("cosign_rekor_url", None),
                    self.target_settings.get("cosign_sbom_skip_verify_rekor", False),
                )
                # Trying to determine if a multiarch attestation has changed by the new push is
                # too complicated. Let's always remove it and replace it by a new one, even if
                # there are no meaningful changes.
                if ml_attestation_exist:
                    LOG.info(
                        f"Multiarch image {dest_ref} already has an attestation. "
                        "It will be replaced by a new attestation containing updated information."
                    )
                    self.delete_existing_attestation(dest_digest_ref, dir_path)

                for attestation_path in tag_attestations:
                    # calling attest multiple times on the same image will append new attestations
                    # which creates JSONL
                    self.cosign_attest_security_manifest(
                        attestation_path,
                        dest_digest_ref,
                        self.target_settings.get("cosign_rekor_url", None),
                        self.target_settings.get("cosign_sbom_skip_upload_rekor", False),
                    )

    def get_source_item_security_manifests(
        self, item: Any, dir_path: str
    ) -> List[DigestSecurityManifest]:
        """
        Get security manifest of a source image.

        Source image has only one architecture, so at most one security manifest will be present.
        Based on testing, source images don't seem to contain security manifests, so failure to
        extract it doesn't result in an error.

        Args:
            item (ContainerPushItem):
                Push item of a source image.
            dir_path (str):
                Directory path where the manifest file should be created.
        Returns [DigestSecurityManifest]:
            Extracted security manifest.
        """
        source_ref = item.metadata["pull_url"]
        digest = self.src_quay_client.get_manifest_digest(source_ref)
        security_manifest_path = os.path.join(
            dir_path, f"security_manifest_source_{uuid.uuid4().hex}.json"
        )

        result = self.cosign_get_security_manifest(source_ref, security_manifest_path)
        if not result:
            LOG.warning(f"Source image {source_ref} doesn't contain a security manifest")
            return []

        return [
            DigestSecurityManifest(digest=digest, security_manifest_path=security_manifest_path)
        ]

    def get_multiarch_item_security_manifests(
        self, item: Any, dir_path: str
    ) -> List[DigestSecurityManifest]:
        """
        Get security manifests of a multiarch image.

        Each architecture has its own security manifest. Older images don't have security manifests,
        so failure to extract them doesn't result in an error.

        Args:
            item (ContainerPushItem):
                Push item of a multiarch image.
            dir_path (str):
                Directory path where the manifest files should be created.
        Returns [DigestSecurityManifest]:
            Extracted security manifests.
        """
        image_manifests = []
        source_ref = item.metadata["pull_url"]
        # Get only repo so that we can attach arch digests to it
        source_repo = source_ref.split(":")[0]
        manifest_list = cast(
            ManifestList,
            self.src_quay_client.get_manifest(source_ref, media_type=QuayClient.MANIFEST_LIST_TYPE),
        )

        for arch in manifest_list["manifests"]:
            image_ref = f"{source_repo}@{arch['digest']}"
            security_manifest_path = os.path.join(
                dir_path,
                f"security_manifest_{arch['platform']['architecture']}_{uuid.uuid4().hex}.json",
            )

            result = self.cosign_get_security_manifest(image_ref, security_manifest_path)
            if not result:
                LOG.warning(
                    f"Image {source_ref} with architecture {arch['platform']['architecture']}"
                    " doesn't contain a security manifest"
                )
            else:
                image_manifests.append(
                    DigestSecurityManifest(
                        digest=arch["digest"], security_manifest_path=security_manifest_path
                    )
                )

        # I wonder if this should be treated as an error
        if image_manifests and len(image_manifests) != len(manifest_list["manifests"]):
            LOG.error(
                f"Only some architectures of multiarch image {source_ref} have a security manifest"
            )

        return image_manifests

    def push_item_security_manifests(self, item: Any) -> None:
        """
        Generate and push container security manifest images of a push item.

        Args:
            item (ContainerPushItem):
                Push item.
        """
        is_source_image = (
            item.metadata["build"].get("extra", {}).get("image", {}).get("sources_for_nvr", False)
        )
        try:
            self.src_quay_client.get_manifest(
                item.metadata["pull_url"], media_type=QuayClient.MANIFEST_LIST_TYPE
            )
        except ManifestTypeError:
            is_multiach_image = False
        else:
            is_multiach_image = True

        with tempfile.TemporaryDirectory(prefix="security_manifest_") as tmp_dir:
            if is_source_image:
                digest_manifests = self.get_source_item_security_manifests(item, tmp_dir)
            elif is_multiach_image:
                digest_manifests = self.get_multiarch_item_security_manifests(item, tmp_dir)
            # if it's something like v2s1, it definitely doesn't have a security manifest
            else:
                LOG.warning(
                    f"Push item {item} contains neither source image nor multiarch image. "
                    "Skipping the push of security manifest images."
                )
                digest_manifests = []

            destination_repos = self.get_destination_repos(item)
            for digest_manifest in digest_manifests:
                self.merge_and_push_security_manifest(
                    item, digest_manifest, destination_repos, tmp_dir
                )

            if is_multiach_image and digest_manifests:
                self.push_manifest_list_security_manifests(item, tmp_dir)

    @log_step("Push container security manifests")
    def push_security_manifests(self) -> None:
        """
        Attest and push security manifest for each push item in parallel.

        Skopeo login is performed before the workflow is started for each push item. Cosign can
        only get the credentials from the default Docker auth config file, so skopeo is used
        to set the correct value. This has to be done at runtime because the credentials are
        target-specific.
        """
        num_thread_security_manifest_push = self.target_settings.get(
            "num_thread_security_manifest_push", 5
        )

        with LocalExecutor() as executor:
            executor.skopeo_login(
                self.quay_host,
                self.target_settings["dest_quay_user"],
                self.target_settings["dest_quay_password"],
            )

        with ThreadPoolExecutor(max_workers=num_thread_security_manifest_push) as executor:
            future_results = [
                executor.submit(self.push_item_security_manifests, item) for item in self.push_items
            ]
            for future in futures.as_completed(future_results):
                if future.exception():
                    raise cast(BaseException, future.exception())  # pragma: no cover
